# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
import unittest
from pathlib import Path
from unittest import TestCase

from aibolit.metrics.external_methods_called.external_methods_called import ExternalMethodsCalled


class ExternalMethodsCalledTest(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_no_external_method_calls(self):
        self.assertEqual(
            ExternalMethodsCalled().value(Path(self.dir_path, 'NoExternalMethodCalls.java')),
            0,
            'Could not calculate how many external methods are called when none'
        )

    @unittest.skip('Not implemented')
    def test_external_method_calls(self):
        self.assertEqual(
            ExternalMethodsCalled().value(Path(self.dir_path, 'ExternalMethodCalls.java')),
            2,
            'Could not calculate how many external method are called when they exist'
        )

    @unittest.skip('Not implemented')
    def test_inner_class_external_method_calls(self):
        self.assertEqual(
            ExternalMethodsCalled().value(
                Path(self.dir_path, 'InnerClassExternalMethodCalls.java')
            ),
            1,
            'Could not calculate how many external methods called when they exist in inner class'
        )

    @unittest.skip('Not implemented')
    def test_double_external_method_calls(self):
        self.assertEqual(
            ExternalMethodsCalled().value(
                Path(self.dir_path, 'DoubleExternalMethodCalls.java')
            ),
            1,
            'Could not calculate external methods called when they are called more than once'
        )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.metrics.cognitiveC.cognitive_c import CognitiveComplexity
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class CognitiveComplexityTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test1(self):
        filepath = self.current_directory / '1.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 7)

    def test2(self):
        filepath = self.current_directory / '2.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 9)

    def test3(self):
        filepath = self.current_directory / '3.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 10)

    def test4(self):
        filepath = self.current_directory / '4.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 6)

    def test5(self):
        filepath = self.current_directory / '5.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 3)

    def test6(self):
        filepath = self.current_directory / '6.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 14)

    def test7(self):
        filepath = self.current_directory / 'recursion.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 6)

    def test8(self):
        filepath = self.current_directory / 'nested.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 9)

    def test9(self):
        filepath = self.current_directory / '7.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 17)

    def test10(self):
        filepath = self.current_directory / '8.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = CognitiveComplexity()
        value = metric.value(ast)
        self.assertEqual(value, 13)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase
from pathlib import Path

from aibolit.__main__ import flatten
from aibolit.ast_framework import AST, ASTNodeType
from aibolit.ast_framework.java_class_decomposition import decompose_java_class
from aibolit.utils.ast_builder import build_ast
from aibolit.metrics.ncss.ncss import NCSSMetric


class JavaClassDecompositionTestSuite(TestCase):
    cur_dir = Path(__file__).absolute().parent

    def test_strong_decomposition(self):
        class_ast = self._get_class_ast(
            "MethodUseOtherMethodExample.java", "MethodUseOtherMethod"
        )
        class_components = decompose_java_class(class_ast, "strong")
        self.assertEqual(len(class_components), 7)

    def test_weak_decomposition(self):
        class_ast = self._get_class_ast(
            "MethodUseOtherMethodExample.java", "MethodUseOtherMethod"
        )
        class_components = decompose_java_class(class_ast, "weak")
        self.assertEqual(len(class_components), 5)

    def _get_class_ast(self, filename: str, class_name: str) -> AST:
        package_ast = AST.build_from_javalang(
            build_ast(Path(__file__).parent.absolute() / filename)
        )
        package_declaration = package_ast.get_root()
        try:
            class_declaration = next(
                class_declaration
                for class_declaration in package_declaration.types
                if class_declaration.name == class_name
            )
            return package_ast.get_subtree(class_declaration)

        except StopIteration:
            raise ValueError(
                f"File '{filename}' does not have top level class '{class_name}'."
            )

    def test_ncss(self):
        test_data_folder = self.cur_dir / "ncss"
        for filepath in test_data_folder.glob("*.java"):
            with self.subTest(f"Testing decomposition of {filepath}"):
                ast = AST.build_from_javalang(build_ast(filepath))

                classes_ast = [
                    ast.get_subtree(node)
                    for node in ast.get_root().types
                    if node.node_type == ASTNodeType.CLASS_DECLARATION
                ]

                ncss_metric = NCSSMetric()

                # NCSS of a class may not be equal to sum of ncss of methods and fields
                # due to presence of nested classes. To bypass it we calculate NCSS only
                # of methods and fields.
                methods_ncss = 0
                fields_ncss = 0
                components_ncss = 0
                components_qty = 0

                for class_ast in classes_ast:
                    class_declaration = class_ast.get_root()

                    for method in class_declaration.methods:
                        methods_ncss += ncss_metric.value(class_ast.get_subtree(method))

                    for field in class_declaration.fields:
                        fields_ncss += ncss_metric.value(class_ast.get_subtree(field))

                    for component in decompose_java_class(class_ast, "strong"):
                        components_qty += 1
                        components_ncss += ncss_metric.value(component)

                # Each component has a CLASS_DECLARATION node.
                # It increase NCSS of each component by 1.
                # To achieve equality we add number of components
                # to the sum of NCSS of just methods and fields.
                self.assertEqual(methods_ncss + fields_ncss + components_qty, components_ncss)

    def __decompose_with_setter_functionality(self, ignore_getters=False, ignore_setters=False):
        file = str(Path(self.cur_dir, 'LottieImageAsset.java'))
        ast = AST.build_from_javalang(build_ast(file))
        classes_ast = [
            ast.get_subtree(node)
            for node in ast.get_root().types
            if node.node_type == ASTNodeType.CLASS_DECLARATION
        ]
        components = list(decompose_java_class(
            classes_ast[0],
            "strong",
            ignore_setters=ignore_setters,
            ignore_getters=ignore_getters))
        function_names = flatten([
            [x.name for x in list(c.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION))]
            for c in components])
        return function_names

    def test_ignore_setters(self):
        function_names = self.__decompose_with_setter_functionality(ignore_setters=True)
        self.assertTrue('setSomething' not in function_names)
        self.assertTrue('setBitmap' not in function_names)

    def test_do_not_ignore_setters(self):
        function_names = self.__decompose_with_setter_functionality(ignore_setters=False)
        self.assertTrue('setSomething' in function_names)
        self.assertTrue('setBitmap' in function_names)

    def test_ignore_getters(self):
        function_names = self.__decompose_with_setter_functionality(ignore_getters=True)
        self.assertTrue('getWidth' not in function_names)

    def test_do_not_ignore_getters(self):
        function_names = self.__decompose_with_setter_functionality(ignore_getters=False)
        self.assertTrue('getWidth' in function_names)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class MethodChainFind:
    """
    Finds chained methods, i.e. foo().bar()
    """

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for node in ast.get_proxy_nodes(ASTNodeType.CLASS_CREATOR,
                                        ASTNodeType.METHOD_INVOCATION,
                                        ASTNodeType.THIS):
            selectors_qty = self._get_selectors_qty(node)
            if selectors_qty > MethodChainFind._allowed_number_of_selectord[node.node_type]:
                lines.append(node.line)

        return lines

    def _get_selectors_qty(self, node: ASTNode) -> int:
        if not hasattr(node, "selectors") or node.selectors is None:
            return 0

        return len(node.selectors)

    _allowed_number_of_selectord = {
        # found node already is a method invocation, so no further invocations are allowed
        ASTNodeType.METHOD_INVOCATION: 0,
        ASTNodeType.THIS: 1,
        # code example: new Object().foo().bar()
        ASTNodeType.CLASS_CREATOR: 1,
    }



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase, skip
from pathlib import Path

from typing import Set

from aibolit.ast_framework import ASTNodeType
from aibolit.ast_framework.java_package import JavaPackage


@skip('JavaPackage class is deprecated')
class JavaPackageTestCase(TestCase):

    def test_java_package_name(self):
        for filename, package_name in JavaPackageTestCase._java_packages_with_names:
            with self.subTest(f'Filename: {filename}'):
                java_package = JavaPackage(Path(__file__).parent.absolute() / filename)
                self.assertEqual(java_package.name, package_name)

    def test_classes(self):
        for filename, flatten_classes in JavaPackageTestCase._java_packages_with_classes:
            with self.subTest(f'Filename: {filename}'):
                java_package = JavaPackage(Path(__file__).parent.absolute() / filename)
                found_class_names: Set[str] = set()
                for class_name, java_class in java_package.java_classes.items():
                    with self.subTest(f'Class name: {class_name}'):
                        found_class_names.add(class_name)
                        java_class_node_types = [
                            java_class.get_type(node) for node in java_class.get_nodes()
                        ]
                        self.assertEqual(
                            java_class_node_types,
                            flatten_classes[class_name]
                        )
                self.assertEqual(found_class_names, flatten_classes.keys())

    _java_packages_with_names = [
        ('SimpleClass.java', '.'),
        ('TwoClasses.java', 'two.classes')
    ]

    _simple_class_preorder_traversal_types = [
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
    ]

    _first_class_preorder_traversal_types = [
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
    ]

    _second_class_preorder_traversal_types = [
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
    ]

    _java_packages_with_classes = [
        ('SimpleClass.java', {'Simple': _simple_class_preorder_traversal_types}),
        ('TwoClasses.java', {
            'First': _first_class_preorder_traversal_types,
            'Second': _second_class_preorder_traversal_types,
        })
    ]



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from pathlib import Path

from aibolit.patterns.if_return_if_detection.if_detection import CountIfReturn
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class CountIfReturnTestCase(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_2nd_level_inside(self):
        filepath = Path(self.dir_path, "1.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = CountIfReturn()
        lines = pattern.value(ast)
        self.assertEqual(lines, [9, 13])

    def test_no_return_inside(self):
        filepath = Path(self.dir_path, "2.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = CountIfReturn()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_nested_one_goodreturn(self):
        filepath = Path(self.dir_path, "3.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = CountIfReturn()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_nested_one_badreturn(self):
        filepath = Path(self.dir_path, "4.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = CountIfReturn()
        lines = pattern.value(ast)
        self.assertEqual(lines, [9])

    def test_withandwithout_returns(self):
        filepath = Path(self.dir_path, "5.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = CountIfReturn()
        lines = pattern.value(ast)
        self.assertEqual(lines, [9, 17, 19, 21])

    def test_nomoreReturn_and_nested(self):
        filepath = Path(self.dir_path, "6.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = CountIfReturn()
        lines = pattern.value(ast)
        self.assertEqual(lines, [13])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from aibolit.ast_framework.ast_node_type import ASTNodeType  # noqa: F401
from aibolit.ast_framework.ast_node import ASTNode  # noqa: F401
from aibolit.ast_framework.ast import AST  # noqa: F401

# register all standard computed fields from 'computed_fields_catalog'
from aibolit.ast_framework.computed_fields_catalog.standard_fields import (
    register_standard_computed_properties,
)

register_standard_computed_properties()



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import TYPE_CHECKING

from cached_property import cached_property  # type: ignore
from deprecated import deprecated  # type: ignore
from networkx import DiGraph  # type: ignore

from aibolit.ast_framework import AST, ASTNodeType

if TYPE_CHECKING:
    from aibolit.ast_framework.java_class import JavaClass


@deprecated("This functionality must be transmitted to ASTNode")
class JavaClassField(AST):
    def __init__(self, tree: DiGraph, root: int, java_class: 'JavaClass'):
        super().__init__(tree, root)
        self._java_class = java_class

    @cached_property
    def name(self) -> str:
        try:
            field_declarator = next(
                self.children_with_type(self.root, ASTNodeType.VARIABLE_DECLARATOR)
            )
            field_name = next(self.children_with_type(field_declarator, ASTNodeType.STRING))
            return self.tree.nodes[field_name]['string']
        except StopIteration:
            raise ValueError("Provided AST does not has 'STRING' node type right under the root")

    @property
    def java_class(self) -> 'JavaClass':
        return self._java_class



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
import re
from typing import List

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class MethodSiblings:
    '''
    Count those methods, which
    start with the same names
    '''
    def _is_method_names_close(self, node: ASTNode, new_node: ASTNode) -> bool:
        splited_name = re.split('([A-Z][^A-Z]*)', node.name)
        return len(splited_name) > 1 and new_node.name.startswith(splited_name[0])

    def value(self, ast: AST) -> List[int]:
        numbers: List[int] = []
        method_nodes = ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION)

        for node in method_nodes:
            for new_node in method_nodes:
                if (node.node_index < new_node.node_index and
                        self._is_method_names_close(node, new_node)):
                    numbers.extend([node.line, new_node.line])
        return sorted(list(set(numbers)))



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from enum import Enum, auto

from typing import List, Set, Optional


class ScopeStatusFlags(Enum):
    ONLY_VARIABLE_DECLARATIONS_PRESENT = auto()
    INSIDE_VARIABLE_DECLARATION_SUBTREE = auto()
    INSIDE_CALLING_SUPER_CLASS_CONSTRUCTOR_SUBTREE = auto()
    INSIDE_ANNOTATION_SUBTREE = auto()


class ScopeStatus:
    _default_scope_status: Set[ScopeStatusFlags] = \
        {ScopeStatusFlags.ONLY_VARIABLE_DECLARATIONS_PRESENT}

    def __init__(self):
        # Copy _default_scope_status to prevent its modification
        self._scope_stack: List[Set[ScopeStatusFlags]] = \
            [ScopeStatus._default_scope_status.copy()]

    def get_status(self) -> Set[ScopeStatusFlags]:
        try:
            return self._scope_stack[-1]
        except IndexError:
            raise RuntimeError("No scopes registered.")

    def add_flag(self, flag: ScopeStatusFlags) -> None:
        try:
            self._scope_stack[-1].add(flag)
        except IndexError:
            raise RuntimeError("No scopes registered.")

    def remove_flag(self, flag: ScopeStatusFlags) -> None:
        try:
            self._scope_stack[-1].discard(flag)
        except IndexError:
            raise RuntimeError("No scopes registered.")

    def enter_new_scope(
        self, new_scope_status: Optional[Set[ScopeStatusFlags]] = None
    ) -> None:
        if new_scope_status is None:
            new_scope_status = ScopeStatus._default_scope_status
        # Copy new_scope_status to prevent its modification
        self._scope_stack.append(new_scope_status.copy())

    def leave_current_scope(self) -> None:
        try:
            self._scope_stack.pop()
        except IndexError:
            raise RuntimeError("No scopes registered.")



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

"""Configuration module for aibolit.

This module defines configuration factories and providers for patterns and metrics,
providing a central registry for all available code analysis tools.
"""
import os
from pathlib import Path

from aibolit.ast_framework import ASTNodeType

from aibolit.metrics.number_variables.numVariables import NumVars as M7
from aibolit.metrics.cognitiveC.cognitive_c import CognitiveComplexity as M4
from aibolit.metrics.entropy.entropy import Entropy as M1
from aibolit.metrics.lcom4.lcom4 import LCOM4 as M5
from aibolit.metrics.max_diameter.max_diameter import MaxDiameter as M6
from aibolit.metrics.ncss.ncss import NCSSMetric as M2
from aibolit.metrics.spaces.SpaceCounter import IndentationCounter as M3
from aibolit.metrics.NumberMethods.NumberMethods import NumberMethods as M8
from aibolit.metrics.RFC.rfc import RFC as M9
from aibolit.metrics.fanout.FanOut import FanOut as M10
from aibolit.patterns.array_as_argument.array_as_argument import ArrayAsArgument as P22
from aibolit.patterns.assert_in_code.assert_in_code import AssertInCode as P1
from aibolit.patterns.assign_null_finder.assign_null_finder import NullAssignment as P28
from aibolit.patterns.classic_setter.classic_setter import ClassicSetter as P2
from aibolit.patterns.empty_rethrow.empty_rethrow import EmptyRethrow as P3
from aibolit.patterns.er_class.er_class import ErClass as P4
from aibolit.patterns.force_type_casting_finder.force_type_casting_finder import \
    ForceTypeCastingFinder as P5
from aibolit.patterns.if_return_if_detection.if_detection import CountIfReturn as P6
from aibolit.patterns.implements_multi.implements_multi import ImplementsMultiFinder as P7
from aibolit.patterns.instanceof.instance_of import InstanceOf as P8
from aibolit.patterns.joined_validation.joined_validation import JoinedValidation as P23
from aibolit.patterns.many_primary_ctors.many_primary_ctors import ManyPrimaryCtors as P9
from aibolit.patterns.method_chaining.method_chaining import MethodChainFind as P10
from aibolit.patterns.multiple_try.multiple_try import MultipleTry as P11
from aibolit.patterns.multiple_while.multiple_while import MultipleWhile as P29
from aibolit.patterns.nested_blocks.nested_blocks import NestedBlocks as P32
from aibolit.patterns.non_final_attribute.non_final_attribute import NonFinalAttribute as P12
from aibolit.patterns.non_final_class.non_final_class import NonFinalClass as P24
from aibolit.patterns.null_check.null_check import NullCheck as P13
from aibolit.patterns.partially_synchronized_methods.partially_synchronized_methods import (
    PartiallySynchronizedMethods as P14
)
from aibolit.patterns.private_static_method.private_static_method import PrivateStaticMethod as P25
from aibolit.patterns.protected_method.protected_method import ProtectedMethod as P30
from aibolit.patterns.public_static_method.public_static_method import PublicStaticMethod as P26
from aibolit.patterns.redundant_catch.redundant_catch import RedundantCatch as P15
from aibolit.patterns.return_null.return_null import ReturnNull as P16
from aibolit.patterns.send_null.send_null import SendNull as P31
from aibolit.patterns.string_concat.string_concat import StringConcatFinder as P17
from aibolit.patterns.supermethod.supermethod import SuperMethod as P18
from aibolit.patterns.hybrid_constructor.hybrid_constructor import HybridConstructor as P19
from aibolit.patterns.var_decl_diff.var_decl_diff import VarDeclarationDistance as P20
from aibolit.patterns.var_middle.var_middle import VarMiddle as P21
from aibolit.patterns.var_siblings.var_siblings import VarSiblings as P27


class Singleton(type):
    """Metaclass for implementing the Singleton pattern."""
    _instances = {}  # type: ignore

    def __call__(cls, *args, **kwargs):
        """Create or return the singleton instance."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(metaclass=Singleton):
    """Central configuration for patterns and metrics discovery."""

    @staticmethod
    def home_aibolit_folder():
        """Get the home folder for aibolit."""
        return os.environ.get('HOME_AIBOLIT') or '/home/jovyan/aibolit'

    @staticmethod
    def folder_to_save_model_data():
        """Get the folder path for saving model data."""
        model_folder = Path(Config().home_aibolit_folder(), 'aibolit', 'binary_files')
        return os.environ.get('SAVE_MODEL_FOLDER') or model_folder

    @staticmethod
    def folder_model_data():
        """Get the folder path for model data."""
        dir_path = os.path.dirname(os.path.realpath(__file__))
        model_file = Path(Path(dir_path), 'binary_files', 'model.pkl')
        return os.environ.get('HOME_MODEL_FOLDER') or model_file

    @staticmethod
    def dataset_file():
        """Get the dataset file path."""
        dataset_path = Path(Config().home_aibolit_folder(), 'scripts', 'target', 'dataset.csv')
        return os.environ.get('HOME_DATASET_CSV') or dataset_path

    @staticmethod
    def train_csv():
        """Get the training CSV file path."""
        return os.environ.get('HOME_TRAIN_DATASET')

    @staticmethod
    def test_csv():
        """Get the test CSV file path."""
        return os.environ.get('HOME_TEST_DATASET')

    @staticmethod
    def get_patterns_config():
        """Get the patterns configuration dictionary."""
        return {
            "patterns": [
                {"name": "Asserts", "code": "P1", "make": P1},
                {"name": "Setters", "code": "P2", "make": P2},
                {"name": "Empty Rethrow", "code": "P3", "make": P3},
                {"name": "Prohibited class name", "code": "P4", "make": P4},
                {"name": "Force Type Casting", "code": "P5", "make": P5},
                {"name": "Count If Return", "code": "P6", "make": P6},
                {"name": "Implements Multi", "code": "P7", "make": P7},
                {"name": "Instance of", "code": "P8", "make": P8},
                {"name": "Many primary constructors", "code": "P9", "make": P9},
                {"name": "Method chain", "code": "P10", "make": P10},
                {"name": "Multiple try", "code": "P11", "make": P11},
                {"name": "Non final attribute", "code": "P12", "make": P12},
                {"name": "Null check", "code": "P13", "make": P13},
                {"name": "Partial synchronized", "code": "P14", "make": P14},
                {"name": "Redundant catch", "code": "P15", "make": P15},
                {"name": "Return null", "code": "P16", "make": P16},
                {"name": "String concat", "code": "P17", "make": P17},
                {"name": "Super Method", "code": "P18", "make": P18},
                {"name": "This in constructor", "code": "P19", "make": P19},
                {
                    "name": "Var declaration distance for 5 lines",
                    "code": "P20_5",
                    "make": lambda: P20(5)
                },
                {
                    "name": "Var declaration distance for 7 lines",
                    "code": "P20_7",
                    "make": lambda: P20(7)
                },
                {
                    "name": "Var declaration distance for 11 lines",
                    "code": "P20_11",
                    "make": lambda: P20(11)
                },
                {"name": "Var in the middle", "code": "P21", "make": P21},
                {"name": "Array as function argument", "code": "P22", "make": P22},
                {"name": "Joined validation", "code": "P23", "make": P23},
                {"name": "Non final class", "code": "P24", "make": P24},
                {"name": "Private static method", "code": "P25", "make": P25},
                {"name": "Public static method", "code": "P26", "make": P26},
                {"name": "Var siblings", "code": "P27", "make": P27},
                {"name": "Null Assignment", "code": "P28", "make": P28},
                {"name": "Multiple While", "code": "P29", "make": P29},
                {"name": "Protected Method", "code": "P30", "make": P30},
                {"name": "Send Null", "code": "P31", "make": P31},
                {"name": "Nested Loop", "code": "P32",
                 "make": lambda: P32(2, ASTNodeType.DO_STATEMENT,
                                     ASTNodeType.FOR_STATEMENT,
                                     ASTNodeType.WHILE_STATEMENT)}
            ],
            "metrics": [
                {"name": "Entropy", "code": "M1", "make": M1},
                {"name": "NCSS lightweight", "code": "M2", "make": M2},
                {
                    "name": "Indentation counter: Right total variance",
                    "code": "M3_1",
                    "make": lambda: M3(right_var=True)
                },
                {
                    "name": "Indentation counter: Left total variance",
                    "code": "M3_2",
                    "make": lambda: M3(left_var=True)
                },
                {
                    "name": "Indentation counter: Right max variance",
                    "code": "M3_3",
                    "make": lambda: M3(max_right=True)
                },
                {
                    "name": "Indentation counter: Left max variance",
                    "code": "M3_4",
                    "make": lambda: M3(max_left=True)
                },
                {"name": "Cognitive Complexity", "code": "M4", "make": M4},
                {"name": "LCOM4", "code": "M5", "make": M5},
                {"name": "Max diameter of AST", "code": "M6", "make": M6},
                {"name": "Number of variables", "code": "M7", "make": M7},
                {"name": "Number of methods", "code": "M8", "make": M8},
                {"name": "Responce for class", "code": "M9", "make": M9},
                {"name": "Fan out", "code": "M10", "make": M10},
            ],
            "target": {

            },
            "patterns_exclude": [
                "P27",  # empty implementation
                "P20_5", "P20_7", "P20_11",  # wasn't refactored yet
                "P28", "P9",  # patterns based on text cannot accept arbitrary AST
            ],
            "metrics_exclude": ["M1", "M3_1", "M3_2", "M3_3", "M3_4", "M5", "M7", "M8"]
        }



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os.path
from pathlib import Path
from unittest import TestCase

from aibolit.patterns.assert_in_code.assert_in_code import AssertInCode
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class AssertInCodeTestCase(TestCase):
    cur_file_dir = Path(os.path.realpath(__file__)).parent

    def test_assert_in_code(self):
        file = Path(self.cur_file_dir, "Book.java")
        ast = AST.build_from_javalang(build_ast(file))
        self.assertEqual(AssertInCode().value(ast), [6])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
import shutil
import subprocess
import tempfile
import uuid

from bs4 import BeautifulSoup


class NPathMetric():
    """Main NPath Complexity class."""

    input = ''

    def __init__(self, input):
        """Initialize class."""
        if len(input) == 0:
            raise ValueError('Empty file for analysis')
        self.input = input

    def value(self, showoutput=False):
        """Run NPath Complexity analaysis"""
        try:
            root = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
            dirName = os.path.join(root, 'src/main/java')
            os.makedirs(dirName)
            if os.path.isdir(self.input):
                shutil.copytree(self.input, os.path.join(dirName, self.input))
            elif os.path.isfile(self.input):
                pos1 = self.input.rfind('/')
                os.makedirs(f'{dirName}/{self.input[0:pos1]}')
                shutil.copyfile(self.input, os.path.join(dirName, self.input))
            else:
                raise Exception(' '.join(['File', self.input, 'does not exist']))
            tmppath = os.path.dirname(os.path.realpath(__file__))
            shutil.copyfile(os.path.join(tmppath, 'pom.xml'), f'{root}/pom.xml')
            shutil.copyfile(os.path.join(tmppath, 'npath.xml'), f'{root}/npath.xml')
            if showoutput:
                subprocess.run(['mvn', 'pmd:pmd'], cwd=root, check=True)
            else:
                subprocess.run(['mvn', 'pmd:pmd'], cwd=root, check=True,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL)
            if os.path.isfile(f'{root}/target/pmd.xml'):
                res = self.__parseFile(root)
                return res
            raise Exception(' '.join(['File', self.input, 'analyze failed']))
        finally:
            shutil.rmtree(root)

    def __parseFile(self, root):
        result = {'data': [], 'errors': []}
        content = []
        with open(f'{root}/target/pmd.xml', 'r', encoding='utf-8') as file:
            content = file.read()
            soup = BeautifulSoup(content, 'lxml')
            files = soup.find_all('file')
            for file in files:
                out = file.violation.string
                name = file['name']
                pos1 = name.find(f'{root}/src/main/java/')
                pos1 = pos1 + len(f'{root}/src/main/java/')
                name = name[pos1:]
                s = 'NPath complexity of '
                pos1 = out.find(s)
                pos1 = pos1 + len(s)
                complexity = int(out[pos1:])
                result['data'].append({'file': name, 'complexity': complexity})
            errors = soup.find_all('error')
            for error in errors:
                name = error['filename']
                pos1 = name.find(f'{root}/src/main/java/')
                pos1 = pos1 + len(f'{root}/src/main/java/')
                name = name[pos1:]
                raise Exception(error['msg'])
        return result



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os.path
from pathlib import Path
from unittest import TestCase
import unittest
from aibolit.patterns.bidirect_index.bidirect_index import BidirectIndex


@unittest.skip("Not implemented")
class BidirectIndexTestCase(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_bidirect_index_increase_decrease(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, "BidirectIndexIncreaseDecrease.java")),
            [3],
            "Could not find bidirect index when index increased and then decreased",
        )

    def test_bidirect_index_decrease_increase(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, "BidirectIndexDecreaseIncrease.java")),
            [3],
            "Could not find bidirect index when index decreased and then increased",
        )

    def test_bidirect_index_increase_decrease_assignment(self):
        self.assertEqual(
            BidirectIndex().value(
                Path(self.dir_path, "BidirectIndexIncreaseDecreaseAssignment.java")
            ),
            [3],
            "Could not find bidirect index when index increased and then decreased with assignment",
        )

    def test_bidirect_index_increase_assignment_decrease(self):
        self.assertEqual(
            BidirectIndex().value(
                Path(self.dir_path, "BidirectIndexIncreaseAssignmentDecrease.java")
            ),
            [3],
            "Could not find bidirect index when index increased with assignment and then decreased",
        )

    def test_bidirect_index_increase_assignment_decrease_assignment(self):
        self.assertEqual(
            BidirectIndex().value(
                Path(self.dir_path, "BidirectIndexIncreaseAssignmentDecreaseAssignment.java")
            ),
            [3],
            "Could not find bidirect index when index increased with assignment "
            "and then decreased with assignment",
        )

    def test_bidirect_index_hidden_scope_true(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, "BidirectIndexHiddenScope.java")),
            [0],
            "Could not find bidirec index when scope is hidden",
        )

    def test_bidirect_index_outsider(self):
        self.assertEqual(
            BidirectIndex().value(Path(self.dir_path, "BidirectIndexOutsider.java")),
            [10],
            "Could not find bidirec index when index is ot of loop",
        )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import subprocess
import os


class HVMetric():
    """Main Halstead Volume class."""

    input = ''

    def __init__(self, input):
        """Initialize class."""
        if len(input) == 0:
            raise ValueError('Empty file for analysis')
        self.input = input

    def value(self):
        """Run Halstead Volume analaysis"""
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..',
                               'halsteadvolume/target/java-project-1.0-SNAPSHOT.jar'))
        if not os.path.isfile(path):
            hvpath = os.path.abspath(os.path.join(os.path.dirname(__file__), '..',
                                                  'halsteadvolume'))
            subprocess.run(['mvn', 'clean', 'package'], cwd=hvpath, check=True)
        result = subprocess.run(['java', '-jar', path, self.input],
                                stdout=subprocess.PIPE, check=True)
        out = result.stdout.decode('utf-8')
        res = result = {'data': [{'file': self.input, 'halsteadvolume': float(out)}]}
        return res



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.classic_setter.classic_setter import ClassicSetter
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class SetterTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_one_valid_patterns(self):
        filepath = self.current_directory / "BaseKeyframeAnimation.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ClassicSetter()
        lines = pattern.value(ast)
        self.assertEqual(lines, [43])

    def test_four_setter_patterns(self):
        filepath = self.current_directory / "Configuration.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ClassicSetter()
        lines = pattern.value(ast)
        self.assertEqual(lines,
                         [905, 909, 1239, 1243, 1364, 1542, 1615, 1644, 1673,
                          1706, 1715, 1726, 1804, 1973, 2023, 2339,
                          2510, 3786, 3822])

    def test_another_setter_patterns(self):
        filepath = self.current_directory / "SequenceFile.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ClassicSetter()
        lines = pattern.value(ast)
        self.assertEqual(lines, [262, 747, 2852, 2858, 2864, 3130])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase

from aibolit.patterns.redundant_catch.redundant_catch import RedundantCatch
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class RedundantCatchTestCase(TestCase):
    def test_simple(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/Simple.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [6])

    def test_both_catches(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/BothCatches.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [6])

    def test_fake(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/TrickyFake.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_try_inside_anonymous(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/TryInsideAnonymous.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [9, 17])

    def test_multiple_catch(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/MultipleCatch.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [6])

    def test_sequential_catch(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/SequentialCatch.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [6])

    def test_sequential_catch_try(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/SequentialCatchTry.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [6, 13])

    def test_try_inside_catch(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/TryInsideCatch.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [10])

    def test_try_inside_finally(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/TryInsideFinally.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [11])

    def test_try_inside_try(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/TryInsideTry.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [8])

    def test_catch_with_functions(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/CatchWithFunctions.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [9])

    def test_catch_with_similar_name(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/NotThrow.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [258])

    def test_try_without_throws(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/ExcelReader.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_try_in_constructor(self):
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/ExcelAnalyserImpl.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [46])

    def test_fake_try_in_lambda(self):
        """
        If function has throws, the pattern shouldn't be recognized
        if the same exception is caught in anonymous lambda
        """
        pattern = RedundantCatch()
        filepath = os.path.dirname(os.path.realpath(__file__)) + "/Cache.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        lines = pattern.value(ast)
        self.assertEqual(lines, [])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.metrics.max_diameter.max_diameter import MaxDiameter
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class MaxDiameterTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test1(self):
        ast = AST.build_from_javalang(build_ast(self.current_directory / '1.java'))
        metric = MaxDiameter()
        metric_value = metric.value(ast)
        self.assertEqual(metric_value, 20)

    def test2(self):
        ast = AST.build_from_javalang(build_ast(self.current_directory / '2.java'))
        metric = MaxDiameter()
        metric_value = metric.value(ast)
        self.assertEqual(metric_value, 14)

    def test3(self):
        ast = AST.build_from_javalang(build_ast(self.current_directory / '3.java'))
        metric = MaxDiameter()
        metric_value = metric.value(ast)
        self.assertEqual(metric_value, 7)

    def test4(self):
        ast = AST.build_from_javalang(build_ast(self.current_directory / '4.java'))
        metric = MaxDiameter()
        metric_value = metric.value(ast)
        self.assertEqual(metric_value, 10)

    def test5(self):
        ast = AST.build_from_javalang(build_ast(self.current_directory / '5.java'))
        metric = MaxDiameter()
        metric_value = metric.value(ast)
        self.assertEqual(metric_value, 11)

    def test6(self):
        ast = AST.build_from_javalang(build_ast(self.current_directory / '6.java'))
        metric = MaxDiameter()
        metric_value = metric.value(ast)
        self.assertEqual(metric_value, 8)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase, skip
from pathlib import Path

from aibolit.ast_framework.java_package import JavaPackage


@skip('JavaClassField is deprecated')
class JavaClassFieldTestCase(TestCase):
    def test_field_name(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'TwoClasses.java')
        java_class = java_package.java_classes['Second']
        self.assertEqual(java_class.fields.keys(), {'x', 'y'})



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import re
from typing import List

from aibolit.utils.encoding_detector import read_text_with_autodetected_encoding


class NullAssignment:
    def value(self, filename: str) -> List:
        source_code = read_text_with_autodetected_encoding(filename)
        pattern = r'[^=!><]=(\s)*null(\s)*;'
        return [lineIndex + 1 for lineIndex, line in
                enumerate(source_code.split('\n')) if re.search(pattern, line)]



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNodeType


class InstanceOf:
    """
    Finds instance_of operator and .isInstance() method call.
    """

    def value(self, ast: AST):
        lines: List[int] = []
        for binary_operator in ast.get_proxy_nodes(ASTNodeType.BINARY_OPERATION):
            if binary_operator.operator == 'instanceof':
                lines.append(binary_operator.line)

        for method_invocation in ast.get_proxy_nodes(ASTNodeType.METHOD_INVOCATION):
            if method_invocation.member == 'isInstance':
                lines.append(method_invocation.line)

        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from aibolit.patterns.assign_null_finder.assign_null_finder import NullAssignment


class NullAssignmentTestCase(TestCase):
    cur_dir = os.path.dirname(os.path.realpath(__file__))

    def test_several(self):
        lines = NullAssignment().value(self.cur_dir + "/several.java")
        self.assertEqual(lines, [8, 9, 14, 18, 24, 25])

    def test_one(self):
        lines = NullAssignment().value(self.cur_dir + "/one.java")
        self.assertEqual(lines, [11])

    def test_not_null(self):
        lines = NullAssignment().value(self.cur_dir + "/not_null.java")
        self.assertEqual(lines, [])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

import javalang
import numpy as np  # type: ignore
from scipy.stats import entropy  # type: ignore

from aibolit.utils.encoding_detector import read_text_with_autodetected_encoding


class Entropy:
    def __init__(self):
        pass

    def __file_to_tokens(self, filename: str) -> List[str]:
        '''Takes path to java class file and returns tokens'''
        source_code = read_text_with_autodetected_encoding(filename)
        tokens = javalang.tokenizer.tokenize(source_code)
        return [token.value for token in tokens]

    def value(self, filename: str):
        tokens = self.__file_to_tokens(filename)
        _, counts = np.unique(tokens, return_counts=True)
        return entropy(counts)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.many_primary_ctors.many_primary_ctors import ManyPrimaryCtors
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class ManyPrimaryCtorsTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_many_primary_ctors(self):
        filepath = self.current_directory / "Book.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ManyPrimaryCtors()
        lines = pattern.value(ast)
        self.assertEqual(lines, [7, 11])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.non_final_attribute.non_final_attribute import NonFinalAttribute
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class NonFinalAttributeTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_find_non_final_attributes(self):
        filepath = self.current_directory / "NonFinalAttribute.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NonFinalAttribute()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5, 7, 9, 11])

    def test_nested_class(self):
        filepath = self.current_directory / "File.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NonFinalAttribute()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15, 19, 67])

    def test_attribute_in_interface(self):
        filepath = self.current_directory / "AttributeInInterface.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NonFinalAttribute()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from javalang.parse import parse
from javalang.tree import CompilationUnit

from aibolit.utils.encoding_detector import read_text_with_autodetected_encoding


def build_ast(filename: str) -> CompilationUnit:
    return parse(read_text_with_autodetected_encoding(filename))



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import unittest

from aibolit.metrics.ncss.ncss import NCSSMetric
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class TestNCSSMetric(unittest.TestCase):
    def testZeroScore(self):
        file = "test/metrics/ncss/Empty.java"
        ast = AST.build_from_javalang(build_ast(file))
        metric = NCSSMetric()
        res = metric.value(ast)
        self.assertEqual(res, 0)

    def testLowScore(self):
        file = "test/metrics/ncss/Simple.java"
        ast = AST.build_from_javalang(build_ast(file))
        metric = NCSSMetric()
        res = metric.value(ast)
        self.assertEqual(res, 2)

    def testBasicExample(self):
        file = "test/metrics/ncss/BasicExample.java"
        ast = AST.build_from_javalang(build_ast(file))
        metric = NCSSMetric()
        res = metric.value(ast)
        self.assertEqual(res, 13)

    def testSimpleExample(self):
        file = "test/metrics/ncss/SimpleExample.java"
        ast = AST.build_from_javalang(build_ast(file))
        metric = NCSSMetric()
        res = metric.value(ast)
        self.assertEqual(res, 18)

    def testSimpleExample2(self):
        file = "test/metrics/ncss/SimpleExample2.java"
        ast = AST.build_from_javalang(build_ast(file))
        metric = NCSSMetric()
        res = metric.value(ast)
        self.assertEqual(res, 19)

    def testChainedIfElse(self):
        file = "test/metrics/ncss/ChainedIfElse.java"
        ast = AST.build_from_javalang(build_ast(file))
        metric = NCSSMetric()
        res = metric.value(ast)
        self.assertEqual(res, 11)

    def testChainedIfElseWithTrailingElse(self):
        file = "test/metrics/ncss/ChainedIfElseWithTrailingElse.java"
        ast = AST.build_from_javalang(build_ast(file))
        metric = NCSSMetric()
        res = metric.value(ast)
        self.assertEqual(res, 12)

    def testFinallyBlock(self):
        file = "test/metrics/ncss/FinallyBlock.java"
        ast = AST.build_from_javalang(build_ast(file))
        metric = NCSSMetric()
        res = metric.value(ast)
        self.assertEqual(res, 7)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.protected_method.protected_method import ProtectedMethod
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class ProtectedMethodTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_not_find_protected_method(self):
        filepath = self.current_directory / "NoProtectedMethod.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ProtectedMethod()
        lines = pattern.value(ast)
        self.assertEqual(lines, [], "Should not match pattern protected method")

    def test_find_protected_method(self):
        filepath = self.current_directory / "ProtectedMethod.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ProtectedMethod()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5, 9], "Should match pattern protected method")

    def test_find_protected_method_inner(self):
        filepath = self.current_directory / "InnerClassProtectedMethod.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ProtectedMethod()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5, 14], "Should match pattern protected method in inner class")

    def test_find_protected_method_anonymous(self):
        filepath = self.current_directory / "AnonymousClassProtectedMethod.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ProtectedMethod()
        lines = pattern.value(ast)
        self.assertEqual(lines, [8], "Should match pattern protected method in anonymous class")



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List, Iterator, Optional

from networkx import DiGraph  # type: ignore

from .ast import AST
from .ast_node import ASTNode
from .ast_node_type import ASTNodeType
from .scope_extractors import extract_scopes


class Scope:
    def __init__(self, scope_tree: DiGraph, scope_id: int):
        self._scope_tree = scope_tree
        self._scope_id = scope_id

    @staticmethod
    def build_from_method_ast(method_ast: AST) -> "Scope":
        method_declaration = method_ast.get_root()
        assert (
            method_declaration.node_type == ASTNodeType.METHOD_DECLARATION
        ), "Building scopes tree supported only for method declaration."

        scope_tree = DiGraph()
        root_scopes_id = Scope._create_scopes_from_node(method_declaration, method_ast, scope_tree)

        assert len(root_scopes_id) == 1, "Method declaration must produce a single scope."
        return Scope(scope_tree, root_scopes_id[0])

    @property
    def parent_scope(self) -> Optional["Scope"]:
        parent_scope_id: Optional[int] = next(self._scope_tree.predecessors(self._scope_id), None)

        if parent_scope_id is None:
            return None

        return Scope(self._scope_tree, parent_scope_id)

    @property
    def nested_scopes(self) -> Iterator["Scope"]:
        for nested_scope_id in self._scope_tree.successors(self._scope_id):
            yield Scope(self._scope_tree, nested_scope_id)

    @property
    def statements(self) -> List[ASTNode]:
        return self._scope_tree.nodes[self._scope_id]["statements"]

    @property
    def parent_node(self) -> ASTNode:
        return self._scope_tree.nodes[self._scope_id]["parent_node"]

    @property
    def parameters(self) -> List[ASTNode]:
        return self._scope_tree.nodes[self._scope_id]["parameters"]

    @staticmethod
    def _create_scopes_from_node(node: ASTNode, method_ast: AST, scope_tree: DiGraph) -> List[int]:
        scopes = extract_scopes(node, method_ast)

        new_scopes_ids: List[int] = []
        for scope in scopes:
            new_scope_id = len(scope_tree)
            new_scopes_ids.append(new_scope_id)
            scope_tree.add_node(
                new_scope_id,
                statements=scope.statements,
                parent_node=scope.parent_node,
                parameters=scope.parameters,
            )

            for statement in scope.statements:
                nested_scopes_ids = Scope._create_scopes_from_node(
                    statement, method_ast, scope_tree
                )
                for nested_scope_id in nested_scopes_ids:
                    scope_tree.add_edge(new_scope_id, nested_scope_id)

        return new_scopes_ids



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List

from aibolit.ast_framework import ASTNodeType, AST


class ErClass:
    '''
    Check if a class name include the forbidden word
    '''
    forbiden_words_in_class_names = (
        'manager',
        'controller',
        'router',
        'dispatcher',
        'printer',
        'writer',
        'reader',
        'parser',
        'generator',
        'renderer',
        'listener',
        'producer',
        'holder',
        'interceptor')

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for node in ast.get_proxy_nodes(ASTNodeType.CLASS_DECLARATION):
            class_name = node.name.lower()
            if any(forbiden_word in class_name
                   for forbiden_word in self.forbiden_words_in_class_names):
                lines.append(node.line)
        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class NestedBlocks:
    '''
    Returns lines in the file where
    nested blocks statements are located
    '''

    def __init__(self, max_depth: int, *block_types: ASTNodeType):
        self._max_depth = max_depth
        self._block_types = block_types

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for block_type in self._block_types:
            for block_ast in ast.get_subtrees(block_type):
                for overnested_block in self._find_overnested_blocks(block_ast.get_root()):
                    lines.append(overnested_block.line)

        return lines

    def _find_overnested_blocks(self, node: ASTNode, depth: int = 1) -> List[ASTNode]:
        if depth == self._max_depth:
            return [node]

        overnested_blocks: List[ASTNode] = []
        for nested_block in self._get_nested_statements(node):
            if nested_block.node_type == node.node_type:
                overnested_blocks.extend(self._find_overnested_blocks(nested_block, depth + 1))

        return overnested_blocks

    def _get_nested_statements(self, node: ASTNode) -> List[ASTNode]:
        nested_statements: List[ASTNode] = []

        if node.node_type == ASTNodeType.IF_STATEMENT:
            nested_statements.extend(self._get_block_statements_list(node.then_statement))

            while (node.else_statement is not None and
                   node.else_statement.node_type == ASTNodeType.IF_STATEMENT):
                node = node.else_statement
                nested_statements.extend(self._get_block_statements_list(node.then_statement))

            if node.else_statement is not None:
                nested_statements.extend(self._get_block_statements_list(node.else_statement))

        elif node.node_type in {ASTNodeType.DO_STATEMENT,
                                ASTNodeType.FOR_STATEMENT,
                                ASTNodeType.WHILE_STATEMENT}:
            nested_statements.extend(self._get_block_statements_list(node.body))

        return nested_statements

    def _get_block_statements_list(self, node: ASTNode) -> List[ASTNode]:
        if node.node_type == ASTNodeType.BLOCK_STATEMENT:
            return node.statements

        # A single statement is treated as a block with this statement
        # This happens in following situations
        # ```java
        # while(True)
        #     doSmth();
        # ```
        return [node]



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class CountIfReturn:
    '''
    Finds "if" statements with else branches and return statement in then branch.
    '''

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for if_statement in ast.get_proxy_nodes(ASTNodeType.IF_STATEMENT):
            if if_statement.else_statement is not None and \
               self._is_then_branch_return(if_statement):
                lines.append(if_statement.line)

        return lines

    def _is_then_branch_return(self, if_statement: ASTNode) -> bool:
        if if_statement.then_statement.node_type == ASTNodeType.RETURN_STATEMENT:
            return True

        if if_statement.then_statement.node_type == ASTNodeType.BLOCK_STATEMENT:
            return any(statement.node_type == ASTNodeType.RETURN_STATEMENT
                       for statement in if_statement.then_statement.statements)

        return False



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
import pickle
from typing import Dict, Any, Tuple
from numpy.typing import NDArray

import numpy as np
import pandas as pd  # type: ignore[import-untyped]
from aibolit.config import Config
from aibolit.model.model import PatternRankingModel, scale_dataset, get_minimum  # noqa: F401 type: ignore


class Stats(object):

    @staticmethod
    def aibolit_stat(test_csv: pd.DataFrame, model=None) -> pd.DataFrame:
        if not model:
            load_model_file = Config.folder_model_data()
            print('Loading model from file {}:'.format(load_model_file))
            with open(load_model_file, 'rb') as fid:
                model = pickle.load(fid)
                print('Model has been loaded successfully')

        scaled_dataset = scale_dataset(test_csv, model.features_conf, "M4")
        cleaned_dataset = scaled_dataset[model.features_conf['features_order'] + ['M2']]
        ranked, _, acts_complexity, acts = Stats.check_impact(
            cleaned_dataset.values,
            model
        )

        m, p = Stats.count_acts(acts.astype(np.int64), ranked.astype(np.int64))
        return Stats.get_table(model.features_conf['features_order'], m, p, acts_complexity)

    @staticmethod
    def count_acts(
            acts: NDArray[np.int64],
            ranked: NDArray[np.int64]) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        patterns_numbers = ranked[:, 0]
        # number of times when pattern was on first place,
        # if we decrease pattern by 1/ncss
        m = np.zeros(ranked.shape[1])
        # number of times when pattern was on first place,
        # if we increase pattern by 1/ncss
        p = np.zeros(ranked.shape[1])
        for i in range(len(patterns_numbers)):
            if acts[i] == 1:
                m[patterns_numbers[i]] += 1
            elif acts[i] == 2:
                p[patterns_numbers[i]] += 1
        return m, p

    @staticmethod
    def get_patterns_name() -> Dict[Any, Any]:
        only_patterns = []
        patterns_code = []
        config = Config.get_patterns_config()
        for x in config['patterns']:
            if x['code'] not in config['patterns_exclude']:
                only_patterns.append(x['name'])
                patterns_code.append(x['code'])
        features_number = len(only_patterns)
        print("Number of features: ", features_number)
        patterns = {x['code']: x['name'] for x in config['patterns']}
        metrics = {x['code']: x['name'] for x in config['metrics']}
        replace_dict = dict(patterns, **metrics)
        return replace_dict

    @staticmethod
    def get_table(
            features_conf: Dict[Any, Any],
            m: NDArray[np.float64],
            p: NDArray[np.float64],
            acts_complexity) -> pd.DataFrame:
        """
        Prints results, given with `check_impact`.


        :param features_conf: features config of model
        :param m: number of times when pattern was on first place,
        if we decrease pattern by 1/ncss
        :param p: number of times when pattern was on first place,
        if we increase pattern by 1/ncss
        :param acts_complexity:

        """

        df = pd.DataFrame(columns=[
            'pattern', ' -1(top1)', '+1(top1)',
            'p-c-', 'p+c+', 'p-c+', 'p+c-', 'p-c=',
            'p+c='])
        replace_dict = Stats.get_patterns_name()
        for i in range(len(features_conf)):
            top_minus = int(m[i])
            top_plus = int(p[i])
            p_minus_c_minus = int(acts_complexity[i, 0])
            p_plus_c_plus = int(acts_complexity[i, 4])
            p_minus_c_plus = int(acts_complexity[i, 1])
            p_plus_c_minus = int(acts_complexity[i, 3])
            p_minus_c_euq = int(acts_complexity[i, 2])
            p_plus_c_euq = int(acts_complexity[i, 5])
            pattern = replace_dict.get(features_conf[i])
            df = pd.concat([df, pd.DataFrame([{
                'pattern': pattern, ' -1(top1)': top_minus, '+1(top1)': top_plus,
                'p-c-': p_minus_c_minus, 'p+c+': p_plus_c_plus, 'p-c+': p_minus_c_plus,
                'p+c-': p_plus_c_minus, 'p-c=': p_minus_c_euq,
                'p+c=': p_plus_c_euq
            }])], ignore_index=True)

        return df

    @staticmethod
    def split_dataset_by_pattern_value(
            X: NDArray[np.float64],
            pattern_idx: int) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        """ Divide dataset.

        :param X: dataset
        :param pattern_idx: pattern index
        :return:
        1st is dataset with pattern where pattern can be null,
        2nd is dataset with pattern where pattern is not null,
        """
        nulls = []
        not_nulls = []
        for snipp in X:
            if snipp[pattern_idx] == 0:
                nulls.append(snipp)
            else:
                not_nulls.append(snipp)

        return np.array(nulls), np.array(not_nulls)

    @staticmethod
    def change_matrix_by_value(
            arr: NDArray[np.float64],
            mask: NDArray[np.bool_],
            i: int,
            incr: NDArray[np.float64]) -> NDArray[np.float64]:
        """
        Args:
            X: np.array with shape (number of snippets, number of patterns).
            mask: np.array with shape (number of snippets, number of patterns).
            i: int, 0 <= i < number of patterns.
            incr: matrix values to add.
            mask: matrix of bools
        Returns:
            X1: modified np.array with shape (number of snippets, number of patterns).
        """

        X1 = arr.copy()
        X1[:, i] += incr[mask[:, i]]

        return X1

    @staticmethod
    def check_impact(
            X: NDArray[np.float64],
            model_input: Any) -> Tuple[NDArray[np.int64], NDArray[np.float64],
                                       NDArray[np.float64], NDArray[np.float64]]:
        """
        Args:
            X: np.array with shape (number of snippets, number of patterns) or
                (number of patterns, ).
            model_input: model to use
        Returns:
            ranked: np.array with shape (number of snippets, number of patterns)
                of sorted patterns in non-increasing order for each snippet of
                code.
            acts: np.array with shape (number of snippets, ) of
            numbers of necessary actions for complexity's decrement.
            0 - do not modify the pattern, 1 - decrease by 1, 2 - increase by 1.
            importances: importances of patters
            acts_complexity: matrix with number of time for each pattern when:
                pattern was increased, complexity increased,
                pattern was increased, complexity decreased, etc.
        """

        if X.ndim == 1:
            X = X.copy()
            X = np.expand_dims(X, axis=0)
        ncss = X[:, -1]
        X = X[:, :-1]

        k = X.shape[1]
        complexity = model_input.model.predict(X)
        importances = np.zeros(X.shape)
        actions = np.zeros(X.shape)
        acts_complexity = np.zeros((X.shape[1], 6))
        for i in range(k):
            nulls, not_nulls = Stats.split_dataset_by_pattern_value(X, i)
            mask: NDArray[np.bool_] = not_nulls > 0
            dec_arr = Stats.change_matrix_by_value(not_nulls, mask, i, -1.0 / ncss[X[:, i] > 0])
            complexity_minus = model_input.model.predict(dec_arr)
            incr_arr = Stats.change_matrix_by_value(not_nulls, mask, i, 1.0 / ncss[X[:, i] > 0])
            complexity_plus = model_input.model.predict(incr_arr)
            c, number = get_minimum(complexity[X[:, i] > 0], complexity_minus, complexity_plus)
            importances[:, i][X[:, i] > 0] = complexity[X[:, i] > 0] - c
            actions[:, i][X[:, i] > 0] = number
            acts_complexity[i, 0] += (complexity_minus < complexity[X[:, i] > 0]).sum()
            acts_complexity[i, 1] += (complexity_minus > complexity[X[:, i] > 0]).sum()
            acts_complexity[i, 2] += (complexity_minus == complexity[X[:, i] > 0]).sum()
            acts_complexity[i, 3] += (complexity_plus < complexity[X[:, i] > 0]).sum()
            acts_complexity[i, 4] += (complexity_plus > complexity[X[:, i] > 0]).sum()
            acts_complexity[i, 5] += (complexity_plus == complexity[X[:, i] > 0]).sum()

        ranked = np.argsort(-1 * importances, 1)
        acts = actions[np.argsort(ranked, 1) == 0]
        return ranked, importances, acts_complexity, acts



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase
from pathlib import Path
from typing import List, Iterator

from aibolit.ast_framework import AST, ASTNodeType
from aibolit.ast_framework.scope import Scope
from aibolit.utils.ast_builder import build_ast

StatementsTypes = List[ASTNodeType]


class ScopeTestCase(TestCase):
    def test_plain_method(self) -> None:
        self._test_method("plain_method")

    def test_lambda_in_assert(self) -> None:
        self._test_method("lambda_in_assert")

    def test_nested_blocks(self) -> None:
        self._test_method("nested_blocks")

    def test_while_cycle(self) -> None:
        self._test_method("while_cycle")

    def test_for_cycle(self) -> None:
        self._test_method("for_cycle")

    def test_if_branching(self) -> None:
        self._test_method("if_branching")

    def test_local_variables(self) -> None:
        self._test_method("local_variables")

    def test_statements_with_expressions(self) -> None:
        self._test_method("statements_with_expressions")

    def test_switch_branching(self) -> None:
        self._test_method("switch_branching")

    def test_synchronized_block(self) -> None:
        self._test_method("synchronized_block")

    def test_try_statement(self) -> None:
        self._test_method("try_statement")

    def test_multiline_lambda(self) -> None:
        self._test_method("multiline_lambda")

    def test_deep_nesting(self) -> None:
        self._test_method("deep_nesting")

    def test_lambda_parameters(self) -> None:
        scope = Scope.build_from_method_ast(self._get_method_ast("multiline_lambda"))
        lambda_scope = next(scope.nested_scopes)

        self.assertEqual(lambda_scope.parent_node.node_type, ASTNodeType.LAMBDA_EXPRESSION)
        self.assertEqual([parameter.name for parameter in lambda_scope.parameters], ["x", "y"])

    def _test_method(self, method_name: str) -> None:
        scope = Scope.build_from_method_ast(self._get_method_ast(method_name))
        self.assertScope(scope, self._scope_statements_in_preorder_by_method[method_name])

    def assertScope(
        self, scope: Scope, statements_by_scope_in_preorder: List[StatementsTypes]
    ) -> None:
        scope_preorder_index = 0
        self.assertStatements(scope, statements_by_scope_in_preorder[scope_preorder_index])

        scope_stack: List[Iterator[Scope]] = [scope.nested_scopes]
        while len(scope_stack) > 0:
            try:
                scope = next(scope_stack[-1])
                scope_preorder_index += 1
                self.assertStatements(scope, statements_by_scope_in_preorder[scope_preorder_index])
                scope_stack.append(scope.nested_scopes)
            except StopIteration:
                scope_stack.pop()

        self.assertEqual(scope_preorder_index + 1, len(statements_by_scope_in_preorder))

    def assertStatements(self, scope: Scope, statements: StatementsTypes) -> None:
        self.assertEqual([statement.node_type for statement in scope.statements], statements)

    def _get_method_ast(self, method_name) -> AST:
        path = str(Path(__file__).absolute().parent / "ScopeTest.java")
        ast = AST.build_from_javalang(build_ast(path))
        package_declaration = ast.get_root()

        assert len(package_declaration.types) == 1 and \
            package_declaration.types[0].node_type == ASTNodeType.CLASS_DECLARATION

        class_declaration = package_declaration.types[0]

        try:
            return next(
                ast.get_subtree(method_declaration)
                for method_declaration in class_declaration.methods
                if method_declaration.name == method_name
            )
        except StopIteration:
            raise ValueError(f"There is no method {method_name} in class {class_declaration.name}")

    _scope_statements_in_preorder_by_method = {
        "plain_method": [
            [ASTNodeType.LOCAL_VARIABLE_DECLARATION, ASTNodeType.LOCAL_VARIABLE_DECLARATION]
        ],
        "lambda_in_assert": [
            [ASTNodeType.LOCAL_VARIABLE_DECLARATION, ASTNodeType.ASSERT_STATEMENT],
            [ASTNodeType.BINARY_OPERATION],
        ],
        "nested_blocks": [
            [ASTNodeType.LOCAL_VARIABLE_DECLARATION, ASTNodeType.BLOCK_STATEMENT],
            [ASTNodeType.LOCAL_VARIABLE_DECLARATION],
        ],
        "while_cycle": [
            [
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.DO_STATEMENT,
                ASTNodeType.WHILE_STATEMENT,
                ASTNodeType.WHILE_STATEMENT,
                ASTNodeType.WHILE_STATEMENT,
            ],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.BINARY_OPERATION],
            [ASTNodeType.STATEMENT_EXPRESSION],
        ],
        "for_cycle": [
            [
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.FOR_STATEMENT,
                ASTNodeType.FOR_STATEMENT,
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.FOR_STATEMENT,
                ASTNodeType.FOR_STATEMENT,
                ASTNodeType.FOR_STATEMENT,
            ],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.BINARY_OPERATION],
            [ASTNodeType.STATEMENT_EXPRESSION],
        ],
        "if_branching": [
            [
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.IF_STATEMENT,
                ASTNodeType.IF_STATEMENT,
                ASTNodeType.IF_STATEMENT,
                ASTNodeType.IF_STATEMENT,
                ASTNodeType.IF_STATEMENT,
                ASTNodeType.IF_STATEMENT,
            ],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.BINARY_OPERATION],
            [ASTNodeType.STATEMENT_EXPRESSION],
        ],
        "local_variables": [
            [
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
            ],
            [ASTNodeType.BINARY_OPERATION],
            [ASTNodeType.BINARY_OPERATION],
        ],
        "statements_with_expressions": [
            [
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.STATEMENT_EXPRESSION,
                ASTNodeType.THROW_STATEMENT,
                ASTNodeType.RETURN_STATEMENT,
            ],
            [ASTNodeType.BINARY_OPERATION],
            [ASTNodeType.MEMBER_REFERENCE],
            [ASTNodeType.BINARY_OPERATION],
        ],
        "switch_branching": [
            [
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.SWITCH_STATEMENT,
                ASTNodeType.SWITCH_STATEMENT,
            ],
            [
                ASTNodeType.STATEMENT_EXPRESSION,
                ASTNodeType.STATEMENT_EXPRESSION,
                ASTNodeType.STATEMENT_EXPRESSION,
            ],
            [ASTNodeType.BINARY_OPERATION],
            [ASTNodeType.STATEMENT_EXPRESSION],
        ],
        "synchronized_block": [
            [
                ASTNodeType.LOCAL_VARIABLE_DECLARATION,
                ASTNodeType.SYNCHRONIZED_STATEMENT,
                ASTNodeType.SYNCHRONIZED_STATEMENT,
            ],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.CLASS_CREATOR],
            [ASTNodeType.STATEMENT_EXPRESSION],
        ],
        "try_statement": [
            [ASTNodeType.TRY_STATEMENT],
            [ASTNodeType.MEMBER_REFERENCE],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION],
        ],
        "multiline_lambda": [
            [ASTNodeType.LOCAL_VARIABLE_DECLARATION, ASTNodeType.STATEMENT_EXPRESSION],
            [ASTNodeType.STATEMENT_EXPRESSION, ASTNodeType.RETURN_STATEMENT],
        ],
        "deep_nesting": [
            [ASTNodeType.FOR_STATEMENT],
            [ASTNodeType.IF_STATEMENT],
            [ASTNodeType.ASSERT_STATEMENT],
            [ASTNodeType.BINARY_OPERATION],
        ],
    }



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNodeType


class PartiallySynchronizedMethods:
    """
    Methods and constructors must have no 'synchronized' statements at all,
    including statements in lambda functions and anonymous class methods,
    or consist of single 'synchronized' statement (fully synchronized)
    """

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for method_ast in ast.get_subtrees(
            ASTNodeType.METHOD_DECLARATION, ASTNodeType.CONSTRUCTOR_DECLARATION
        ):
            method_declaration = method_ast.get_root()
            is_fully_sync_method = len(method_declaration.body) == 1 and \
                method_declaration.body[0].node_type == ASTNodeType.SYNCHRONIZED_STATEMENT

            synchronized_statements = list(
                method_ast.get_proxy_nodes(ASTNodeType.SYNCHRONIZED_STATEMENT)
            )

            if len(synchronized_statements) > 0 and not is_fully_sync_method:
                lines.extend(statement.line for statement in synchronized_statements)

        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from os import listdir
from pathlib import Path
from typing import Set

import numpy as np
from tqdm import tqdm

from aibolit.config import Config
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast

# TO-FIX: fix all errors in the patterns/metrics and make these lists empty
EXCLUDE_PATTERNS: Set[str] = {}
EXCLUDE_METRICS: Set[str] = {}

# TO-FIX: refactor or delete following patterns and metrics
PATTERNS_ACCEPT_FILE_PATH = {
    "P20_5",
    "P20_7",
    "P20_11",  # P20* wasn't refactored yet
    "P28",  # patterns based on text cannot accept arbitrary AST
}
METRICS_ACCEPT_FILE_PATH = {"M1", "M3_1", "M3_2", "M3_3", "M3_4", "M5", "M7"}

samples_path = Path(__file__).absolute().parent / "samples"


def _check_pattern(p_info, filepath):
    try:
        pattern = p_info["make"]()
        if p_info["code"] in PATTERNS_ACCEPT_FILE_PATH:
            pattern_result = pattern.value(filepath)
        else:
            ast = AST.build_from_javalang(build_ast(filepath))
            pattern_result = pattern.value(ast)

        assert isinstance(pattern_result, list) and all(
            isinstance(item, int) for item in pattern_result
        ), (
            f"Pattern return {pattern_result} with type {type(pattern_result)}, "
            "but list of int was expected."
        )
    except Exception as exception:
        raise RuntimeError(
            f"Error searching for pattern {p_info['name']} "
            f"with code {p_info['code']} for file {filepath}"
        ) from exception


def _check_metric(m_info, filepath):
    try:
        metric = m_info["make"]()
        if m_info["code"] in METRICS_ACCEPT_FILE_PATH:
            metric_result = metric.value(filepath)
        else:
            ast = AST.build_from_javalang(build_ast(filepath))
            metric_result = metric.value(ast)

        assert isinstance(metric_result, (int, float, np.float64)), (
            f"Metric return {metric_result} of type {type(metric_result)}, "
            "but int, float or numpy float was expected"
        )
    except Exception as exception:
        raise RuntimeError(
            f"Error in application of the metric {m_info['name']} "
            f"with code {m_info['code']} for file {filepath}"
        ) from exception


if __name__ == "__main__":
    config = Config.get_patterns_config()

    print(f"Processed files in {samples_path}:")
    for filename in tqdm(listdir(samples_path)):
        for pattern_info in config["patterns"]:
            if pattern_info["code"] not in EXCLUDE_PATTERNS:
                _check_pattern(pattern_info, samples_path / filename)

        for metric_info in config["metrics"]:
            if metric_info["code"] not in EXCLUDE_METRICS:
                _check_metric(metric_info, samples_path / filename)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import networkx as nx  # type: ignore
from aibolit.utils.ast_builder import build_ast

from aibolit.utils.cohesiongraph import CohesionGraph


class LCOM4:

    coh = CohesionGraph()

    def __init__(self):
        pass

    def value(self, filename: str):

        tree = build_ast(filename)
        G = self.coh.value(tree)
        return nx.number_connected_components(G)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from pathlib import Path

from aibolit.patterns.string_concat.string_concat import StringConcatFinder
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class ConcatStringTestCase(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_concat_strings_in_print(self):
        filepath = Path(self.dir_path, "ConcatInPrint.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [17])

    def test_member_plus_string(self):
        filepath = Path(self.dir_path, "MemberPlusString.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [101, 113])

    def test_empty_case(self):
        filepath = Path(self.dir_path, "Nothing.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_string_plus_member(self):
        filepath = Path(self.dir_path, "StringPlusMember.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 1)

    def test_many_concats(self):
        filepath = Path(self.dir_path, "ManyConcats.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15, 16, 17, 18])

    def test_concat_in_different_methods(self):
        filepath = Path(self.dir_path, "DifferentMethods.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [16, 30])

    def test_fake_operator_plus(self):
        filepath = Path(self.dir_path, "FakePlusOperator.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_string_with_quotes(self):
        filepath = Path(self.dir_path, "RustServerCodegen.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(
            lines,
            [
                354,
                374,
                375,
                381,
                382,
                410,
                411,
                420,
                424,
                431,
                432,
                440,
                458,
                468,
                481,
                496,
                501,
                511,
                564,
                565,
                606,
                607,
                612,
                703,
                997,
                1011,
                1014,
                1031,
                1034,
                1082,
                1190,
                1260,
                1269,
                1339,
                1368,
                1374,
                1388,
                1416,
            ],
        )

    def test_comment_inside_line(self):
        filepath = Path(self.dir_path, "XMLDataObject.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [162, 167, 179, 194, 220, 281])

    def test_fake1(self):
        filepath = Path(self.dir_path, "Chain.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = StringConcatFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [34])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
import re
from typing import List, Dict

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class VarSiblings:
    '''
    Find those variables, which have complex
    names and start with the same word
    '''
    def _collect_method_variables_names(self, ast: AST, node: ASTNode) -> Dict[str, int]:
        assert node.node_type == ASTNodeType.METHOD_DECLARATION
        vars_lines: Dict[str, int] = {}
        for local_var_node in ast.get_proxy_nodes(ASTNodeType.LOCAL_VARIABLE_DECLARATION):
            for var_declaration in local_var_node.declarators:
                var_name = var_declaration.name
                if var_name not in vars_lines:
                    # to filter not complex names
                    temp_name = re.split('([A-Z][^A-Z]*)', var_name)
                    if len(temp_name) > 1:
                        vars_lines[var_name] = local_var_node.line
        return vars_lines

    def _is_names_close(self, node_name_1: str, node_name_2: str) -> bool:
        # here we want to find those names, which start with the same word
        # and the lenght of this word > 3
        common_names_prefix = re.split('([A-Z][^A-Z]*)', node_name_1)[0]
        return node_name_2.startswith(common_names_prefix) and len(common_names_prefix) > 3

    def _find_sibling_vars(self, vars_info: Dict) -> List[int]:
        lines: List[int] = []
        for node_name1 in vars_info:
            for node_name2 in vars_info:
                if (node_name1 != node_name2 and
                        self._is_names_close(node_name1, node_name2) and
                        vars_info[node_name2] not in lines):
                    lines.append(vars_info[node_name2])
        return lines

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for method_declaration in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION):
            variables_info = self._collect_method_variables_names(ast, method_declaration)
            lines.extend(self._find_sibling_vars(variables_info))
        return sorted(lines)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List

from aibolit.ast_framework import ASTNodeType, AST


class ProtectedMethod:
    '''
    Once we find a protected method in a class, it's a pattern.
    '''
    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for method_declaration in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION):
            if 'protected' in method_declaration.modifiers:
                lines.append(method_declaration.line)
        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.return_null.return_null import ReturnNull
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class ReturnNullPatternTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_anonymous(self):
        filepath = self.current_directory / "Anonymous.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ReturnNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [27, 31])

    def test_empty(self):
        filepath = self.current_directory / "Empty.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ReturnNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_with_ternary1(self):
        filepath = self.current_directory / "With_Ternary1.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ReturnNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15])

    def test_with_ternary2(self):
        filepath = self.current_directory / "With_Ternary2.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ReturnNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15])

    def test_with_ternary_not_return_null(self):
        filepath = self.current_directory / "With_Ternary_not_return_null.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ReturnNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple(self):
        filepath = self.current_directory / "Simple.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ReturnNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from pathlib import Path

from aibolit.patterns.er_class.er_class import ErClass
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class ErClassTestCase(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_manager_in_middle(self):
        filepath = Path(self.dir_path, "AnimatableSplitDimensionPathValue.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15])

    def test_controller_in_end(self):
        filepath = Path(self.dir_path, "AnimatableTransform.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15])

    def test_one_normal_class(self):
        filepath = Path(self.dir_path, "AuditEventModelProcessor.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_two_classes_with_pattern(self):
        filepath = Path(self.dir_path, "BaseKeyframeAnimation.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [21, 189])

    def test_class_parser(self):
        filepath = Path(self.dir_path, "Configuration.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [3109])

    def test_another_normal_class(self):
        filepath = Path(self.dir_path, "FillContent.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_four_normal_classes(self):
        filepath = Path(self.dir_path, "FJIterateTest.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_two_distant_normal_classes(self):
        filepath = Path(self.dir_path, "FJListProcedureRunner.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_classes_in_comments(self):
        filepath = Path(self.dir_path, "KeyProviderCryptoExtension.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_classes_in_methods(self):
        filepath = Path(self.dir_path, "OsSecureRandom.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_normal_class(self):
        filepath = Path(self.dir_path, "RectangleContent.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_three_writers_one_reader(self):
        filepath = Path(self.dir_path, "SequenceFile.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ErClass()
        lines = pattern.value(ast)
        self.assertEqual(lines, [840, 1481, 1541, 1687])



#!/usr/bin/env python
# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

"""The main entry point. Invoke as `aibolit' or `python -m aibolit'.
"""

import argparse
import concurrent.futures
import json
import multiprocessing
import operator
import os
import pickle
import sys
import time
import traceback
from collections import defaultdict, OrderedDict
from os import scandir
from pathlib import Path
from sys import stdout
from typing import List, Any, Dict, Tuple

import javalang
from javalang.parser import JavaSyntaxError
import numpy as np  # type: ignore
import requests  # type: ignore[import-untyped]
from lxml import etree  # type: ignore
from packaging.version import parse as parse_version

from aibolit import __version__
from aibolit.ast_framework import AST, ASTNodeType
from aibolit.ast_framework.java_class_decomposition import decompose_java_class
from aibolit.config import Config
from aibolit.metrics.ncss.ncss import NCSSMetric
from aibolit.ml_pipeline.ml_pipeline import train_process, collect_dataset
from aibolit.utils.ast_builder import build_ast

dir_path = os.path.dirname(os.path.realpath(__file__))


def list_dir(path, files):
    dir_list = []
    for entry in scandir(path):
        if entry.is_dir():
            dir_list.append(entry.path)
            dir_list.extend(list_dir(entry.path, files))
        elif entry.is_file() \
                and entry.path.endswith('.java') \
                and entry.name != 'package-info.java':
            files.append(entry.path)
    return dir_list


def run_parse_args(commands_dict):
    parser = argparse.ArgumentParser(
        description='Find the pattern which has the largest impact on readability',
        usage='''aibolit <command> [<args>]

        You can run 1 command:
        train          Train model
        check          Recommend pattern
        recommend      Recommend pattern. The same as recommend, just another acronym''')

    parser.add_argument('command', help='Subcommand to run')
    parser.add_argument(
        '--version', action='version',
        version=f'%(prog)s {__version__}'
    )
    # parse_args defaults to [1:] for args, but you need to
    # exclude the rest of the args too, or validation will fail
    args = parser.parse_args(sys.argv[1:2])
    if args.command not in commands_dict:
        parser.print_help()
        exit(1)

    return commands_dict[args.command]()


def train():
    parser = argparse.ArgumentParser(
        description='Collect dataset and train model'
    )
    parser.add_argument(
        '--java_folder',
        help='full path to directory',
        default=False,
        required=False
    )
    parser.add_argument(
        '--max_classes',
        type=lambda v: sys.maxsize if v == '' else int(v),
        required=False,
        default=None
    )
    parser.add_argument(
        '--skip_collect_dataset',
        required=False,
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--dataset_file',
        required=False,
        default=False
    )
    parser.add_argument(
        "--target-metric",
        choices=["rfc", "fanout", "diameter", "cognitive"],
        default="cognitive",
        help="Select target metric. Available metrics: rfc, fanout, diameter, cognitive. "
             "Default is cognitive."
    )
    args = parser.parse_args(sys.argv[2:])
    if not args.skip_collect_dataset:
        collect_dataset(args)

    metric_name_to_code = {
        "rfc": "M9",
        "fanout": "M10",
        "diameter": "M6",
        "cognitive": "M4",
    }
    train_process(metric_name_to_code[args.target_metric])


def __count_value(value_dict, input_params, code_lines_dict, java_file: str, is_metric=False):
    """
    Count value for input dict

    :param value_dict: Pattern item or Metric item from CONFIG
    :param input_params: list with calculated patterns/metrics
    :param code_lines_dict: list with found code lines of patterns/metrics
    :param java_file: full path for java file
    :is_metric: is item metric?
    :return: None, it has side-effect
    """
    acronym = value_dict['code']
    try:
        ast = AST.build_from_javalang(build_ast(java_file))
        val = value_dict['make']().value(ast)
        if not is_metric:
            input_params[acronym] = len(val)
            code_lines_dict['lines_' + acronym] = val
        else:
            input_params[acronym] = val
    except Exception:
        exc_type, exc_value, exc_tb = sys.exc_info()
        raise Exception(f"Can't count {acronym} metric: {str(type(exc_value))}")


def flatten(lst):
    return [item for sublist in lst for item in sublist]


def add_pattern_if_ignored(
        dct: Dict[str, Any],
        pattern_item: Dict[Any, Any],
        results_list: List[Any]) -> None:
    """ If pattern code is not ignored, add it to the result list

    :param dct: dict, where key is pattern, value is list of lines range to ignore
    :param pattern_item: pattern dict which was get after `inference` function
    :param results_list: result list to add

    """
    ignored_lines = dct.get(pattern_item['pattern_code'])
    if ignored_lines:
        for place in ignored_lines:
            # get lines range of ignored code
            start_line_to_ignore = place[0]
            end_line_to_ignore = place[1]
            new_code_lines = []
            for line in pattern_item['code_lines']:
                if start_line_to_ignore <= line <= end_line_to_ignore:
                    continue
                else:
                    new_code_lines.append(line)
            pattern_item['code_lines'] = new_code_lines
            if len(pattern_item['code_lines']) > 0:
                results_list.append(pattern_item)
    else:
        results_list.append(pattern_item)


def find_annotation_by_node_type(
        tree: javalang.tree.CompilationUnit,
        node_type) -> Dict[Any, Any]:
    """Search nodes with annotations.

    :param tree: javalang.tree
    :param node_type: Node type of javalang.tree
    :return
    dict with annotations, where key is node, value is list of string annotations;
    """
    annonations = defaultdict(list)
    for _, node in tree.filter(node_type):
        if node.annotations:
            for a in node.annotations:
                if hasattr(a.element, 'value'):
                    if 'aibolit' in a.element.value:
                        annonations[node].append(
                            a.element.value.split('.')[1].rstrip('\"')
                        )
                elif hasattr(a.element, 'values'):
                    for j in a.element.values:
                        if 'aibolit' in j.value:
                            annonations[node].append(
                                j.value.split('.')[1].rstrip('\"')
                            )
    return annonations


def find_start_and_end_lines(node) -> Tuple[int, int]:  # noqa: C901
    max_line = node.position.line

    def check_max_position(node):
        nonlocal max_line
        if hasattr(node, '_position'):
            if node.position.line > max_line:
                max_line = node.position.line

    def traverse(node):
        check_max_position(node)

        if hasattr(node, 'children'):
            for child in node.children:
                if isinstance(child, list) and (len(child) > 0):
                    for item in child:
                        if isinstance(item, list):
                            for i in item:
                                traverse(i)
                        else:
                            traverse(item)
                else:
                    if hasattr(child, 'children'):
                        infants = child.children
                        for infant in infants:
                            if isinstance(infant, list):
                                for j in infant:
                                    traverse(j)
                            else:
                                traverse(infant)
                    else:
                        check_max_position(child)

    traverse(node)
    return node.position.line, max_line


def calculate_patterns_and_metrics_with_decomposition(
        file_path: str,
        args):
    error_exc = None
    patterns_to_suppress = args.suppress
    results_for_components = []

    try:
        config = Config.get_patterns_config()
        patterns_info = [
            pattern_info
            for pattern_info in config["patterns"]
            if pattern_info["code"] not in config["patterns_exclude"]
        ]

        metrics_info = [
            metric_info
            for metric_info in config["metrics"]
            if metric_info["code"] not in config["metrics_exclude"]
        ]
        ast = AST.build_from_javalang(build_ast(file_path))
        classes_ast = [
            ast.get_subtree(node)
            for node in ast.get_root().types
            if node.node_type == ASTNodeType.CLASS_DECLARATION
        ]
        for class_ast in classes_ast:
            for index, component_ast in enumerate(decompose_java_class(class_ast, "strong")):
                result_for_component: Dict[Any, Any] = {}
                code_lines_dict: Dict[Any, Any] = OrderedDict()
                input_params = OrderedDict()  # type: ignore

                for pattern_info in patterns_info:
                    if pattern_info['code'] in config['patterns_exclude']:
                        continue
                    if pattern_info['code'] in patterns_to_suppress:
                        input_params[pattern_info["code"]] = 0
                        code_lines_dict["lines_" + pattern_info["code"]] = []
                    else:
                        pattern = pattern_info["make"]()
                        pattern_result = pattern.value(component_ast)
                        input_params[pattern_info["code"]] = len(pattern_result)
                        code_lines_dict["lines_" + pattern_info["code"]] = pattern_result

                for metric_info in metrics_info:
                    metric = metric_info["make"]()
                    metric_result = metric.value(component_ast)
                    input_params[metric_info["code"]] = metric_result

                result_for_component['code_lines_dict'] = code_lines_dict
                result_for_component['input_params'] = input_params
                result_for_component['index'] = index
                results_for_components.append(result_for_component)
    except Exception as ex:
        error_exc = ex

    return results_for_components, error_exc


def calculate_patterns_and_metrics(file, args):
    code_lines_dict = input_params = {}  # type: ignore
    error_exc = None
    patterns_to_suppress = args.suppress
    try:
        config = Config.get_patterns_config()
        for pattern in config['patterns']:
            if pattern['code'] in config['patterns_exclude']:
                continue
            if pattern['code'] in patterns_to_suppress:
                input_params[pattern['code']] = 0
            else:
                __count_value(pattern, input_params, code_lines_dict, file)

        for metric in config['metrics']:
            if metric['code'] in config['metrics_exclude']:
                continue
            __count_value(metric, input_params, code_lines_dict, file, is_metric=True)
    except Exception as ex:
        error_exc = ex
        input_params = []  # type: ignore

    return input_params, code_lines_dict, error_exc


def inference(
        input_params: Dict[Any, Any],
        code_lines_dict,
        args):
    """
    Find a pattern which has the largest impact on target

    :param input_params: list if calculated patterns/metrics
    :param code_lines_dict: list with found code lines of patterns/metrics
    :param file: filename

    :return: list of results with code_lies for each pattern and its name
    """
    model_path = args.model
    do_full_report = args.full
    results = []
    if input_params:
        if not model_path:
            model_path = Config.folder_model_data()
        with open(model_path, 'rb') as fid:
            model = pickle.load(fid)
        sorted_result, importances = model.predict(input_params)
        patterns_list = model.features_conf['features_order']
        for iter, (key, val) in enumerate(sorted_result.items()):
            if key in patterns_list:
                pattern_code = key
                code_lines = code_lines_dict.get('lines_' + key)
                importance = importances[iter] * input_params[pattern_code]

                # We show only patterns with positive importance
                if code_lines and importance > 0:
                    if code_lines:
                        pattern_name = \
                            [x['name'] for x in Config.get_patterns_config()['patterns']
                             if x['code'] == pattern_code][0]
                        results.append(
                            {'code_lines': code_lines,
                             'pattern_code': pattern_code,
                             'pattern_name': pattern_name,
                             'importance': importance
                             })
                    if not do_full_report:
                        break
    else:
        return results

    return results


def create_results(
        input_params: Dict[Any, Any],
        code_lines_dict: Dict[Any, Any],
        args: argparse.Namespace,
        classes_with_patterns_ignored: List[Any],
        patterns_ignored: Dict[Any, Any]):

    results_list = inference(input_params, code_lines_dict, args)
    new_results: List[Any] = []
    for pattern_item in results_list:
        # check if the whole class is suppressed
        if pattern_item['pattern_code'] not in classes_with_patterns_ignored:
            # then check if patterns are ignored in fields or functions
            add_pattern_if_ignored(patterns_ignored, pattern_item, new_results)
            # add_pattern_if_ignored(patterns_for_fields_ignored, pattern_item, new_results)
        else:
            continue

    return new_results


def _extract_patterns_ignored(tree):
    """Extract patterns that should be ignored from annotations."""
    classes_with_annonations = find_annotation_by_node_type(
        tree, javalang.tree.ClassDeclaration)
    functions_with_annotations = find_annotation_by_node_type(
        tree, javalang.tree.MethodDeclaration)
    fields_with_annotations = find_annotation_by_node_type(
        tree, javalang.tree.FieldDeclaration)
    classes_with_patterns_ignored = flatten(
        [pattern_code for node, pattern_code in classes_with_annonations.items()])
    patterns_ignored = defaultdict(list)

    for node, patterns_list in functions_with_annotations.items():
        start_pos, end_pos = find_start_and_end_lines(node)
        for p in patterns_list:
            patterns_ignored[p].append([start_pos, end_pos])

    for node, patterns_list in fields_with_annotations.items():
        for p in patterns_list:
            patterns_ignored[p].append([node.position.line,
                                       node.position.line])

    return classes_with_patterns_ignored, patterns_ignored


def _process_components(components, args, classes_with_patterns_ignored, patterns_ignored):
    """Process components to create results."""
    results_list = []
    for lcom_component in components:
        code_lines_dict = lcom_component.get('code_lines_dict')
        input_params = lcom_component.get('input_params')
        ranked_results = create_results(
            input_params,
            code_lines_dict,
            args,
            classes_with_patterns_ignored,
            patterns_ignored
        )
        if ranked_results:
            results_list.append(ranked_results)
    return results_list


def run_recommend_for_file(file: str, args):  # flake8: noqa
    """
    Calculate patterns and metrics, pass values to model and suggest pattern to change
    :param file: file to analyze
    :param args: different command line arguments
    :return: dict with code lines, filename and pattern name
    """
    java_file = str(Path(os.getcwd(), file))
    ncss = 0
    try:
        tree = build_ast(file)
        classes_with_patterns_ignored, patterns_ignored = _extract_patterns_ignored(tree)

        components, error_exception = calculate_patterns_and_metrics_with_decomposition(
            java_file, args)

        if not components:
            results_list = []  # type: ignore
            error_exception = 'Empty java file; ncss = 0'
        else:
            results_list = _process_components(
                components, args, classes_with_patterns_ignored, patterns_ignored)

        ncss = NCSSMetric().value(AST.build_from_javalang(build_ast(file)))
    except Exception as e:
        error_exception = e
        ncss = 0
        results_list = []

    return {
        'filename': file,
        'results': results_list,
        'exception': error_exception,
        'ncss': ncss,
    }


def _process_pattern(patterns_tag, pattern, i, patterns_number):
    """Process a single pattern and add it to the XML."""
    pattern_item = etree.SubElement(patterns_tag, 'pattern')
    pattern_name_str = pattern.get('pattern_name')
    details = etree.SubElement(pattern_item, 'details')
    details.text = pattern_name_str or ''
    pattern_item.attrib['code'] = pattern.get('pattern_code')
    code_lines_items = pattern.get('code_lines')
    pattern_score = pattern.get('importance')
    pattern_score_tag = etree.SubElement(pattern_item, 'score')
    pattern_score_tag.text = f'{pattern_score:.2f}' or ''
    pattern_score_tag = etree.SubElement(pattern_item, 'order')
    pattern_score_tag.text = f'{i}/{patterns_number}'

    if code_lines_items:
        code_lines_lst_tree_node = etree.SubElement(pattern_item, 'lines')
        for code_line in code_lines_items:
            code_line_elem = etree.SubElement(code_lines_lst_tree_node, 'number')
            code_line_elem.text = str(code_line)

    return pattern_score


def _process_file_result(file, result_for_file):
    """Process a single file result and add it to the XML."""
    filename = result_for_file.get('filename')
    name = etree.SubElement(file, 'path')
    output_string_tag = etree.SubElement(file, 'summary')
    name.text = filename
    results_item = result_for_file.get('results')
    ex = result_for_file.get('exception')
    errors_string = str(result_for_file.get('exception')) or type(ex).__name__

    if not results_item and not errors_string:
        output_string = 'Your code is perfect in aibolit\'s opinion'
        output_string_tag.text = output_string
        return None
    elif not results_item and errors_string:
        output_string = f'Error when calculating patterns: {str(errors_string)}'
        output_string_tag.text = output_string
        return None
    else:
        output_string = 'Some issues found'
        output_string_tag.text = output_string
        importances_sum_tag = etree.SubElement(file, 'score')
        patterns_tag = etree.SubElement(file, 'patterns')
        patterns_number = len(result_for_file['results'])
        importance_for_class = []

        for i, pattern in enumerate(result_for_file['results'], start=1):
            if pattern.get('pattern_code'):
                pattern_score = _process_pattern(patterns_tag, pattern, i, patterns_number)
                importance_for_class.append(pattern_score)

        score_for_class = sum(importance_for_class)
        importances_sum_tag.text = str(score_for_class)
        return score_for_class


def _create_xml_header(top, results, cmd, exit_code):
    """Create XML header with metadata."""
    header_tag = etree.SubElement(top, 'header')
    importances_for_all_classes_tag = etree.SubElement(header_tag, 'score')
    datetime_tag = etree.SubElement(header_tag, 'datetime')
    datetime_tag.addprevious(etree.Comment('datetime in ms'))
    datetime_tag.text = str(int(round(time.time() * 1000)))
    version_tag = etree.SubElement(header_tag, 'version')
    version_tag.text = str(__version__)
    files_number_tag = etree.SubElement(header_tag, 'files')
    files_number_tag.addprevious(etree.Comment('Files with found patterns'))
    files_number_tag.text = str(len(
        [x for x in results
         if (not x.get('exception') and x.get('results'))]
    ))
    patterns_number_tag = etree.SubElement(header_tag, 'patterns')
    ncss_tag = etree.SubElement(header_tag, 'ncss')
    ncss_tag.text = str(sum([x['ncss'] for x in results]))
    cmd_tag = etree.SubElement(header_tag, 'cmd')
    cmd_tag.text = ' '.join(cmd)
    cmd_tag = etree.SubElement(header_tag, 'exit_code')
    cmd_tag.text = str(exit_code)
    return importances_for_all_classes_tag, patterns_number_tag


def create_xml_tree(results, full_report, cmd, exit_code):
    """
    Creates xml from output of `check` function
    :param results: output of `check` function
    :return: xml string
    """
    importances_for_all_classes = []
    top = etree.Element('report')
    importances_tag, patterns_number_tag = _create_xml_header(top, results, cmd, exit_code)

    total_patterns = 0
    files = etree.SubElement(top, 'files')
    if not full_report:
        files.addprevious(etree.Comment(
            'Show pattern with the largest contribution to Cognitive Complexity'))
    else:
        files.addprevious(etree.Comment('Show all patterns'))
    for result_for_file in results:
        file = etree.SubElement(files, 'file')
        score_for_file = _process_file_result(file, result_for_file)
        if score_for_file is not None:
            total_patterns += len(result_for_file['results'])
            importances_for_all_classes.append(score_for_file)

    patterns_number_tag.text = str(total_patterns)
    if importances_for_all_classes:
        importances_tag.text = str(np.mean(importances_for_all_classes))

    return top


def get_exit_code(results):
    """
    Analyzed recommendation results and generate exit_code for pipeline
    """

    files_analyzed = len(results)
    errors_number = 0
    perfect_code_number = 0
    errors_strings = []
    for result_for_file in results:
        results = result_for_file.get('results')
        ex = result_for_file.get('exception')
        if not results and not ex:
            perfect_code_number += 1
        elif not results and ex:
            if not isinstance(ex, JavaSyntaxError):
                errors_strings.append(ex)
                errors_number += 1
            else:
                # ignore JavaSyntaxError, it is expected error
                perfect_code_number += 1

    if len(errors_strings) == files_analyzed:
        # we have errors everywhere
        exit_code = 2
    elif perfect_code_number == files_analyzed:
        # everything is good
        exit_code = 0
    else:
        # we have some recommendation
        exit_code = 1

    return exit_code


def create_text(results, full_report, is_long=False):
    importances_for_all_classes = []
    buffer = []
    total_patterns = 0
    if not full_report:
        buffer.append('Show pattern with the largest contribution to Cognitive Complexity')
    else:
        buffer.append('Show all patterns')
    for result_for_file in results:
        filename = result_for_file.get('filename')
        results_item = result_for_file.get('results')
        ex = result_for_file.get('exception')
        if not results_item and not ex:
            # Do nothing, patterns were not found
            pass
        if not results_item and ex:
            output_string = (f'{filename}: error when calculating patterns: '
                             f'{str(ex) or type(ex).__name__}')
            buffer.append(output_string)
        elif results_item and not ex:
            # get unique patterns score
            pattern_number = 0
            cur_pattern_name = ''
            for pattern_item in result_for_file['results']:
                pattern_score = pattern_item.get('importance')
                code = pattern_item.get('pattern_code')
                if code:
                    pattern_name_str = pattern_item.get('pattern_name')
                    if cur_pattern_name != pattern_name_str:
                        pattern_number += 1
                        cur_pattern_name = pattern_name_str
                    buffer.append(f'{filename}[{pattern_item.get("code_line")}]: '
                                  f'{pattern_name_str} ({code}: {pattern_score:.2f})')

            total_patterns += pattern_number
    if importances_for_all_classes:
        show_summary(buffer, importances_for_all_classes, is_long, results, total_patterns)

    return buffer


def show_summary(buffer, importances_for_all_classes, is_long, results, total_patterns):
    files_number = len(
        [x for x in results
         if (not x.get('errors_string') and x.get('results'))
         ])
    ncss = sum(
        [x['ncss'] for x in results
         if (not x.get('errors_string') and x.get('results'))])
    if not is_long:
        buffer.append(f'Total score: {np.mean(importances_for_all_classes):.2f}, '
                      f'files seen: {files_number}, patterns found: {total_patterns}, '
                      f'ncss: {ncss}')
    else:
        buffer.append(f'Total score: {np.mean(importances_for_all_classes):.2f}, '
                      f'files seen: {files_number}, ncss {ncss}')
        url = 'https://github.com/cqfn/aibolit/blob/master/PATTERNS.md'
        buffer.append(f'You can find all information about patterns here: {url}')


def print_total_score_for_file(
        buffer: List[str],
        filename: str,
        importances_for_all_classes: List[int],
        result_for_file):
    patterns_scores = {}
    for x in result_for_file['results']:
        patterns_scores[x['pattern_name']] = x['importance']
    importances_for_class = sum(patterns_scores.values())
    importances_for_all_classes.append(importances_for_class)
    buffer.append(f'{filename} score: {importances_for_class}')
    return patterns_scores


def check():
    """Run check pipeline."""

    parser = argparse.ArgumentParser(
        description='Get recommendations for Java code',
        usage='aibolit check < --folder | --filenames > [--model] '
              '[--threshold] [--full] [--format]')

    group_exclusive = parser.add_mutually_exclusive_group(required=True)

    group_exclusive.add_argument(
        '--folder',
        help='path to Java files',
        default=False
    )
    group_exclusive.add_argument(
        '--filenames',
        help='list of Java files',
        nargs="*",
        default=False
    )
    parser.add_argument(
        '--model',
        help='''file where pretrained model is located, the default path is located
        in site-packages and is installed with aibolit automatically''',
        default=False
    )
    parser.add_argument(
        '--full',
        help='show all recommendations instead of the best one',
        default=False,
        action='store_true'
    )
    parser.add_argument(
        '--format',
        default='compact',
        help='compact (by default), long or xml. Usage: --format=xml'
    )

    parser.add_argument(
        '--suppress',
        default=[],
        help='Suppress certain patterns (comma separated value)'
    )
    parser.add_argument(
        '--exclude',
        action='append',
        nargs='+',
        default=[],
        help='Glob patterns to ignore. '
    )
    args = parser.parse_args(sys.argv[2:])

    if args.suppress:
        args.suppress = args.suppress.strip().split(',')

    files_to_exclude = handle_exclude_command_line(args)

    files = []
    if args.filenames:
        files = [str(Path(x).absolute()) for x in args.filenames if x not in files_to_exclude]
    elif args.folder:
        all_files = []
        list_dir(args.folder, all_files)
        files = [str(Path(x).absolute()) for x in all_files if str(x) not in files_to_exclude]

    results = list(run_thread(files, args))
    exit_code = get_exit_code(results)

    if args.format:
        if args.format == 'xml':
            root = create_xml_tree(results, args.full, sys.argv, exit_code)
            tree = root.getroottree()
            tree.write(stdout.buffer, pretty_print=True)
        else:
            if args.format in ['compact', 'short']:
                new_results = format_converter_for_pattern(results)
            elif args.format == 'long':
                new_results = format_converter_for_pattern(results, 'code_line')
            else:
                raise Exception('Unknown format')

            text = create_text(new_results, args.full, True)

            if args.format == 'short':
                print(text[-1])
            else:
                print('\n'.join(text))

    return exit_code


def handle_exclude_command_line(args: Any) -> List[str]:
    files_to_exclude: List[str] = []
    full_path_to_exclude = args.folder
    glob_patterns = args.exclude
    for glob_p in glob_patterns:
        pattern = glob_p[0]
        files_to_exclude.extend([str(x.absolute())
                                for x in Path(full_path_to_exclude).glob(pattern)])
    print("ignore:", files_to_exclude)
    return files_to_exclude


def format_converter_for_pattern(results, sorted_by=None):
    """
    Reformat data where data are sorted by patterns importance
    (it is already sorted in the input).
    Then lines are sorted in ascending order.
    """
    total_patterns_list = []
    for file in results:
        components = file.get('results')
        all_results = {}
        if components:
            for i, component in enumerate(components):
                new_items = flatten([
                    [{'pattern_code': x['pattern_code'],
                      'pattern_name': x['pattern_name'],
                      'code_line': line,
                      'importance': x['importance']
                      } for line in sorted(x['code_lines'])] for x in component
                ])
                if not sorted_by:
                    all_results[i] = sorted(
                        new_items,
                        key=lambda e: (-e['importance'], e['pattern_code'], e['code_line'])
                    )
                else:
                    all_results[i] = new_items

        if not sorted_by:
            # iterate over all components, get 1st importance for 1st component,
            # then 1st importance for 2nd component, etc.
            while len(all_results) > 0:
                for idx in list(all_results):
                    componet_res = all_results[idx]
                    top_pattern = componet_res.pop()
                    total_patterns_list.append(top_pattern)
                    if not componet_res:
                        del all_results[idx]
        else:
            for _, val in all_results.items():
                total_patterns_list.extend(val)
            total_patterns_list = sorted(total_patterns_list,
                                         key=operator.itemgetter(sorted_by, 'pattern_code'))

        file['results'] = total_patterns_list

    return results


def version():
    """
    Parses arguments and shows current version of program.
    """

    parser = argparse.ArgumentParser(
        description='Show version')
    parser.add_argument(
        '--version',
    )
    print(f'%(prog)s {__version__}')


def run_thread(files, args):
    """
    Parallel patterns/metrics calculation
    :param files: list of java files to analyze

    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=multiprocessing.cpu_count()) as executor:
        future_results = [executor.submit(run_recommend_for_file, file, args) for file in files]
        concurrent.futures.wait(future_results)
        for future in future_results:
            yield future.result()


def get_versions(pkg_name):
    url = f'https://pypi.python.org/pypi/{pkg_name}/json'
    releases = json.loads(requests.get(url, timeout=15).content)['releases']
    return sorted(releases, key=parse_version, reverse=True)


def main():
    try:
        max_available_version = get_versions('aibolit')[0]
        if max_available_version != __version__:
            print(f'Version {max_available_version} is available, but you are using {__version__}')
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout,
            requests.exceptions.ReadTimeout):
        print('Can\'t check aibolit version. Network is not available')
    try:
        commands = {
            'train': train,
            'check': check,
            'recommend': check,
            'version': version,
        }
        exit_code = run_parse_args(commands)
    except Exception:
        traceback.print_exc()
        sys.exit(2)
    else:
        sys.exit(exit_code)


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.multiple_try.multiple_try import MultipleTry
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class MultipleTryPatternTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_simple(self):
        filepath = self.current_directory / "Simple.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleTry()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5])

    def test_large_file(self):
        filepath = self.current_directory / "Large.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleTry()
        lines = pattern.value(ast)
        self.assertEqual(lines, [622, 708])

    def test_try_inside_anonymous(self):
        filepath = self.current_directory / "TryInsideAnomymous.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleTry()
        lines = pattern.value(ast)
        self.assertEqual(lines, [8])

    def test_try_inside_catch(self):
        filepath = self.current_directory / "TryInsideCatch.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleTry()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5])

    def test_try_inside_finally(self):
        filepath = self.current_directory / "TryInsideFinaly.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleTry()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5])

    def test_try_inside_try(self):
        filepath = self.current_directory / "TryInsideTry.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleTry()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5])

    def test_try_method_overloading(self):
        filepath = self.current_directory / "TryMethodOverloading.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleTry()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework.ast import AST, ASTNodeType


class SuperMethod:
    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for statement in ast.get_proxy_nodes(ASTNodeType.STATEMENT_EXPRESSION):
            if any(child.node_type == ASTNodeType.SUPER_METHOD_INVOCATION
                    for child in statement.children):
                lines.append(statement.line)
        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List, Tuple

from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class MaxDiameter:
    """
    Max diameter of class methods.
    """

    def value(self, ast: AST) -> int:
        method_diameters: List[int] = [
            self._calcalute_diameter(method_ast)
            for method_ast in ast.get_subtrees(ASTNodeType.METHOD_DECLARATION)
        ]

        return max(method_diameters, default=0)

    def _calcalute_diameter(self, ast: AST) -> int:
        distant_node_from_root, _ = self._find_distant_node(ast, ast.get_root(), False)

        # traverse undirected graph, because we need to be able to traverse from child to parent in
        # general. It's not needed at previous call, as the most distant node of a tree is anyway
        # a child of root
        # and there is no need to traverse from child to parent, which simply safe time
        _, diameter = self._find_distant_node(ast, distant_node_from_root, True)
        return diameter

    def _find_distant_node(self, ast: AST, source_node: ASTNode,
                           undirected: bool) -> Tuple[ASTNode, int]:
        distance = 0

        max_distance = 0
        distant_node = source_node

        def on_node_entering(node: ASTNode) -> None:
            nonlocal distance
            distance += 1

        def on_node_leaving(node: ASTNode) -> None:
            nonlocal distance, max_distance, distant_node
            if distance > max_distance:
                max_distance = distance
                distant_node = node

            distance -= 1

        ast.traverse(on_node_entering, on_node_leaving, source_node, undirected)

        return (distant_node, max_distance)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import Any, List

from aibolit.ast_framework import AST, ASTNodeType, ASTNode


class HybridConstructor:

    def is_statement_ctor_inv(self, node: ASTNode) -> bool:
        """Is statement explicit constructor invocation."""

        return node.expression.node_type == ASTNodeType.EXPLICIT_CONSTRUCTOR_INVOCATION

    def traverse_in_if(
            self,
            val: ASTNode,
            exp_ctrs_decls: List[ASTNode],
            other_statements: List[ASTNode]) -> None:
        """Traverse over if condition recursively to find
        explicit constructor invocation."""
        if hasattr(val, 'statements'):
            children = list(val.statements)
            for i in children:
                self.traverse(i, exp_ctrs_decls, other_statements)

        if hasattr(val, 'then_statement'):
            self.traverse_in_if(val.then_statement, exp_ctrs_decls, other_statements)
            other_statements.append(val.then_statement)
        if hasattr(val, 'else_statement'):
            self.traverse_in_if(val.else_statement, exp_ctrs_decls, other_statements)
            other_statements.append(val.else_statement)

    def traverse(
            self,
            statement: ASTNode,
            exp_ctrs_decls: List[ASTNode],
            other_statements: List[ASTNode]) -> None:
        """Traverse over AST recursively to find all explicit
        constructor invocations and other statements."""

        if statement.node_type == ASTNodeType.STATEMENT_EXPRESSION:
            is_ctor_inv = self.is_statement_ctor_inv(statement)
            if is_ctor_inv:
                exp_ctrs_decls.append(statement)
            else:
                other_statements.append(statement)
        elif statement.node_type == ASTNodeType.TRY_STATEMENT:
            self.traverse_in_try(exp_ctrs_decls, other_statements, statement)
        elif statement.node_type in (
                ASTNodeType.DO_STATEMENT,
                ASTNodeType.WHILE_STATEMENT):
            for i in statement.body.children:
                self.traverse(i, exp_ctrs_decls, other_statements)
        elif statement.node_type == ASTNodeType.FOR_STATEMENT:
            for i in statement.body.children:
                other_statements.append(statement)
                self.traverse(i, exp_ctrs_decls, other_statements)
        elif statement.node_type == ASTNodeType.IF_STATEMENT:
            other_statements.append(statement)
            self.traverse_in_if(statement.then_statement, exp_ctrs_decls, other_statements)
            self.traverse_in_if(statement.else_statement, exp_ctrs_decls, other_statements)
        else:
            other_statements.append(statement)

    def traverse_in_try(
            self,
            exp_ctrs_decls: List[ASTNode],
            other_statements: List[ASTNode],
            statement: ASTNode) -> None:
        """Check try statements and find different statements."""
        if (statement.resources is not None) or \
                (statement.catches is not None and statement.catches[0].block != []) or \
                (statement.finally_block is not None):
            other_statements.append(statement)
        for try_stat in statement.block:
            self.traverse(try_stat, exp_ctrs_decls, other_statements)

    def value(self, ast: AST) -> List[int]:
        lines = []
        for node in ast.get_proxy_nodes(ASTNodeType.CONSTRUCTOR_DECLARATION):
            exp_ctrs_decls: List[Any] = []
            other_statements: List[Any] = []
            for statement in node.body:
                self.traverse(statement, exp_ctrs_decls, other_statements)

            if len(exp_ctrs_decls) > 0:
                if len(other_statements) > 0:
                    lines.append(node.line)

        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNodeType


class MultipleWhile:
    '''
    Finds methods, with more than one "while" cycle, exceluding nested ones.
    '''

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for method_declaration in ast.get_subtrees(ASTNodeType.METHOD_DECLARATION):
            top_level_while_loops = method_declaration.get_subtrees(ASTNodeType.WHILE_STATEMENT)
            if len(list(top_level_while_loops)) > 1:
                lines.append(method_declaration.get_root().line)

        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class RedundantCatch:
    '''
    To check wether the method throws same as it does inside the
    try -> catch structure in this method
    '''

    def _is_redundant(self, method_throw_name: List[str], try_node: ASTNode) -> bool:
        assert try_node.node_type == ASTNodeType.TRY_STATEMENT
        for catch_node in try_node.catches:
            for catch_node_name in catch_node.parameter.types:
                if catch_node_name in method_throw_name:
                    return True
        return False

    def _get_lambda_try_nodes(self, ast: AST, lambda_node: ASTNode) -> List[int]:
        assert lambda_node.node_type == ASTNodeType.LAMBDA_EXPRESSION
        return [try_node.line for try_node in
                ast.get_subtree(lambda_node).get_proxy_nodes(ASTNodeType.TRY_STATEMENT)]

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        excluded_nodes: List[int] = []

        for method_declaration in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION,
                                                      ASTNodeType.CONSTRUCTOR_DECLARATION):
            method_throw_names = method_declaration.throws
            for try_node in ast.get_subtree(method_declaration).get_proxy_nodes(
                    ASTNodeType.TRY_STATEMENT):
                if method_throw_names is not None and \
                        try_node.catches is not None and \
                        self._is_redundant(method_throw_names, try_node):
                    lines.append(try_node.line)

            for lambda_node in ast.get_subtree(method_declaration).get_proxy_nodes(
                    ASTNodeType.LAMBDA_EXPRESSION):
                excluded_nodes.extend(self._get_lambda_try_nodes(ast, lambda_node))
        return sorted(list(set(lines).difference(set(excluded_nodes))))



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase

from aibolit.patterns.force_type_casting_finder import force_type_casting_finder
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class ForceTypeCastingFinderTestCase(TestCase):
    def test_simple(self):
        pattern = force_type_casting_finder.ForceTypeCastingFinder()
        path = os.path.dirname(os.path.realpath(__file__)) + "/1.java"
        ast = AST.build_from_javalang(build_ast(path))
        lines = pattern.value(ast)
        self.assertEqual(lines, [8])

    def test_several_casts(self):
        pattern = force_type_casting_finder.ForceTypeCastingFinder()
        path = os.path.dirname(os.path.realpath(__file__)) + "/2.java"
        ast = AST.build_from_javalang(build_ast(path))
        lines = pattern.value(ast)
        self.assertEqual(lines, [8, 14, 20])

    def test_zero_lines(self):
        pattern = force_type_casting_finder.ForceTypeCastingFinder()
        path = os.path.dirname(os.path.realpath(__file__)) + "/3.java"
        ast = AST.build_from_javalang(build_ast(path))
        lines = pattern.value(ast)
        self.assertEqual(lines, [])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
import unittest

from aibolit.patterns.method_siblings.method_siblings import MethodSiblings
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class MethodSiblingsTestCase(unittest.TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_find_simple_method_siblings(self):
        filepath = self.current_directory / "SimpleMethodSiblings.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodSiblings()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5, 8])

    def test_find_alternate_method_siblings(self):
        filepath = self.current_directory / "AlternateMethodSiblings.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodSiblings()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5, 11])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNodeType


class NonFinalClass:
    '''
    Find all classes that are neither "final" nor "abstract".
    '''

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for class_declaration in ast.get_proxy_nodes(ASTNodeType.CLASS_DECLARATION):
            if len(class_declaration.modifiers & NonFinalClass._allowed_class_modifiers) == 0:
                lines.append(class_declaration.line)

        return lines

    _allowed_class_modifiers = {'final', 'abstract'}



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase
from pathlib import Path

from aibolit.ast_framework.java_package import JavaPackage


class CFGBuilderTestCase(TestCase):

    def test_cfg_of_method(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'SimpleClass.java')
        java_class = java_package.java_classes['Simple']
        methods = java_class.methods['Increment']
        self.assertEqual(len(methods), 1)
        method = next(iter(methods))
        self.assertEqual(method.cfg.size(), 2)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework.ast import AST, ASTNodeType


class ImplementsMultiFinder:
    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for node in ast.get_proxy_nodes(ASTNodeType.CLASS_DECLARATION):
            if node.implements and len(node.implements) > 1:
                lines.append(node.line)
        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class EmptyRethrow:
    '''
    Check if we throw the same exception as it was caught
    '''
    def _process_catch(self, ast: AST, catch_clauses: List[ASTNode]):
        lines: List[int] = []
        for catch_clause in catch_clauses:
            throw_statements = ast.get_subtree(catch_clause).get_proxy_nodes(
                ASTNodeType.THROW_STATEMENT)
            for throw_stat in throw_statements:
                if throw_stat.expression.node_type == ASTNodeType.MEMBER_REFERENCE \
                   and throw_stat.expression.member == catch_clause.parameter.name:
                    lines.append(throw_stat.line)
        return lines

    def value(self, ast: AST) -> List[int]:
        total_code_lines: List[int] = []
        for try_node in ast.get_proxy_nodes(ASTNodeType.TRY_STATEMENT):
            catch_clauses = try_node.catches
            if catch_clauses:
                total_code_lines.extend(self._process_catch(ast, catch_clauses))

        return sorted(total_code_lines)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import Union, Any, Callable
from itertools import chain

from aibolit.ast_framework import ASTNode


def chain_field_getter_factory(*steps: Union[str, int]) -> Callable[[ASTNode], Any]:
    """
    A chained field is a field of some other field and so on.
    For example name of class field is retrieved like `field_decl.declarators[0].name`
    using javalang fields. To automate this chain fields can be used.
    Chain field is specified by sequence of strings (field names) and integers (list indexes).
    If a field is list of ASTNodes and next step in chain is string
    then we get that field from every node in list.
    If we get list of lists, it gets flatten.
    List with single items is unwrapped.
    """

    def get_chain_field(node: ASTNode) -> Any:
        field = node
        for step in steps:
            if isinstance(field, list) and \
               isinstance(step, str) and \
               all(isinstance(item, ASTNode) for item in field):
                # get attribute from all elements from a list
                field = [getattr(item, step) for item in field]

                if all(isinstance(item, list) for item in field):  # flattening list
                    field = list(chain.from_iterable(field))

            elif isinstance(field, list) and isinstance(step, int):
                field = field[step]
            elif isinstance(field, ASTNode) and isinstance(step, str):
                field = getattr(field, step)
            else:
                raise RuntimeError(f"Failed to apply step {step} to field {field}.")

        return field

    return get_chain_field



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from aibolit.patterns.var_decl_diff.var_decl_diff import VarDeclarationDistance


class VarDeclarationDiffTestCase(TestCase):
    def test_good_class(self):
        pattern = VarDeclarationDistance(lines_th=2)
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + "/1.java")
        self.assertEqual(lines, [])

    def test_bad_class(self):
        pattern = VarDeclarationDistance(lines_th=2)
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + "/2.java")
        self.assertEqual(lines, [16])

    def test_bad_class2(self):
        pattern = VarDeclarationDistance(lines_th=5)
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + "/3.java")
        self.assertEqual(sorted(lines), [216, 785, 971])

    def test_case_with_multiline_function_arguments(self):
        pattern = VarDeclarationDistance(lines_th=2)
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + "/4.java")
        self.assertEqual(lines, [17, 21])
